#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Methods for computing bivariate normal probabilities and statistics.

.. [Genz2004bvnt]
    A. Genz. Numerical computation of rectangular bivariate and trivariate normal and
    t probabilities. Statistics and Computing, 2004.

.. [Muthen1990moments]
    B. Muthen. Moments of the censored and truncated bivariate normal distribution.
    British Journal of Mathematical and Statistical Psychology, 1990.
"""

from __future__ import annotations

from math import pi as _pi

import torch
from botorch.exceptions import UnsupportedError
from botorch.utils.probability.utils import (
    case_dispatcher,
    get_constants_like,
    leggauss,
    ndtr as Phi,
    phi,
    STANDARDIZED_RANGE,
)
from botorch.utils.safe_math import (
    div as safe_div,
    exp as safe_exp,
    mul as safe_mul,
    sub as safe_sub,
)
from torch import Tensor

# Some useful constants
_inf = float("inf")
_2pi = 2 * _pi
_sqrt_2pi = _2pi**0.5
_inv_2pi = 1 / _2pi


def bvn(r: Tensor, xl: Tensor, yl: Tensor, xu: Tensor, yu: Tensor) -> Tensor:
    r"""A function for computing bivariate normal probabilities.

    Calculates `P(xl < x < xu, yl < y < yu)` where `x` and `y` are bivariate normal with
    unit variance and correlation coefficient `r`. See Section 2.4 of [Genz2004bvnt]_.

    This method uses a sign flip trick to improve numerical performance. Many of `bvnu`s
    internal branches rely on evaluations `Phi(-bound)`. For `a < b < 0`, the term
    `Phi(-a) - Phi(-b)` goes to zero faster than `Phi(b) - Phi(a)` because
    `finfo(dtype).epsneg` is typically much larger than `finfo(dtype).tiny`. In these
    cases, flipping the sign can prevent situations where `bvnu(...) - bvnu(...)` would
    otherwise be zero due to round-off error.

    Args:
        r: Tensor of correlation coefficients.
        xl: Tensor of lower bounds for `x`, same shape as `r`.
        yl: Tensor of lower bounds for `y`, same shape as `r`.
        xu: Tensor of upper bounds for `x`, same shape as `r`.
        yu: Tensor of upper bounds for `y`, same shape as `r`.

    Returns:
        Tensor of probabilities `P(xl < x < xu, yl < y < yu)`.

    """
    if not (r.shape == xl.shape == xu.shape == yl.shape == yu.shape):
        raise UnsupportedError("Arguments to `bvn` must have the same shape.")

    # Sign flip trick
    _0, _1, _2 = get_constants_like(values=(0, 1, 2), ref=r)
    flip_x = xl.abs() > xu  # is xl more negative than xu is positive?
    flip_y = yl.abs() > yu
    flip = (flip_x & (~flip_y | yu.isinf())) | (flip_y & (~flip_x | xu.isinf()))
    if flip.any():  # symmetric calls to `bvnu` below makes swapping bounds unnecessary
        sign = _1 - _2 * flip.to(dtype=r.dtype)
        xl = sign * xl  # becomes `-xu` if flipped
        xu = sign * xu  # becomes `-xl`
        yl = sign * yl  # becomes `-yu`
        yu = sign * yu  # becomes `-yl`

    p = bvnu(r, xl, yl) - bvnu(r, xu, yl) - bvnu(r, xl, yu) + bvnu(r, xu, yu)
    return p.clip(_0, _1)


def bvnu(r: Tensor, h: Tensor, k: Tensor) -> Tensor:
    r"""Solves for `P(x > h, y > k)` where `x` and `y` are standard bivariate normal
    random variables with correlation coefficient `r`. In [Genz2004bvnt]_, this is (1)

        `L(h, k, r) = P(x < -h, y < -k) \
        = 1/(a 2\pi) \int_{h}^{\infty} \int_{k}^{\infty} f(x, y, r) dy dx,`

    where `f(x, y, r) = e^{-1/(2a^2) (x^2 - 2rxy + y^2)}` and `a = (1 - r^2)^{1/2}`.

    [Genz2004bvnt]_ report the following integation scheme incurs a maximum of 5e-16
    error when run in double precision: if `|r| >= 0.925`, use a 20-point quadrature
    rule on a 5th order Taylor expansion; else, numerically integrate in polar
    coordinates using no more than 20 quadrature points.

    Args:
        r: Tensor of correlation coefficients.
        h: Tensor of negative upper bounds for `x`, same shape as `r`.
        k: Tensor of negative upper bounds for `y`, same shape as `r`.

    Returns:
        A tensor of probabilities `P(x > h, y > k)`.
    """
    if not (r.shape == h.shape == k.shape):
        raise UnsupportedError("Arguments to `bvnu` must have the same shape.")
    _0, _1, lower, upper = get_constants_like((0, 1) + STANDARDIZED_RANGE, r)
    x_free = h < lower
    y_free = k < lower
    return case_dispatcher(
        out=torch.empty_like(r),
        cases=(  # Special cases admitting closed-form solutions
            (lambda: (h > upper) | (k > upper), lambda mask: _0),
            (lambda: x_free & y_free, lambda mask: _1),
            (lambda: x_free, lambda mask: Phi(-k[mask])),
            (lambda: y_free, lambda mask: Phi(-h[mask])),
            (lambda: r == _0, lambda mask: Phi(-h[mask]) * Phi(-k[mask])),
            (  # For |r| >= 0.925, use a Taylor approximation
                lambda: r.abs() >= get_constants_like(0.925, r),
                lambda m: _bvnu_taylor(r[m], h[m], k[m]),
            ),
        ),  # For |r| < 0.925, integrate in polar coordinates.
        default=lambda mask: _bvnu_polar(r[mask], h[mask], k[mask]),
    )


def _bvnu_polar(
    r: Tensor, h: Tensor, k: Tensor, num_points: int | None = None
) -> Tensor:
    r"""Solves for `P(x > h, y > k)` by integrating in polar coordinates as

        `L(h, k, r) = \Phi(-h)\Phi(-k) + 1/(2\pi) \int_{0}^{sin^{-1}(r)} f(t) dt \
        f(t) = e^{-0.5 cos(t)^{-2} (h^2 + k^2 - 2hk sin(t))}`

    For details, see Section 2.2 of [Genz2004bvnt]_.
    """
    if num_points is None:
        mar = r.abs().max()
        num_points = 6 if mar < 0.3 else 12 if mar < 0.75 else 20

    _0, _1, _i2, _i2pi = get_constants_like(values=(0, 1, 0.5, _inv_2pi), ref=r)

    x, w = leggauss(num_points, dtype=r.dtype, device=r.device)
    x = x + _1
    asin_r = _i2 * torch.asin(r)
    sin_asrx = (asin_r.unsqueeze(-1) * x).sin()

    _h = h.unsqueeze(-1)
    _k = k.unsqueeze(-1)
    vals = safe_exp(
        safe_sub(safe_mul(sin_asrx, _h * _k), _i2 * (_h.square() + _k.square()))
        / (_1 - sin_asrx.square())
    )
    probs = Phi(-h) * Phi(-k) + _i2pi * asin_r * (vals @ w)
    return probs.clip(min=_0, max=_1)  # necessary due to "safe" handling of inf


def _bvnu_taylor(r: Tensor, h: Tensor, k: Tensor, num_points: int = 20) -> Tensor:
    r"""Solves for `P(x > h, y > k)` via Taylor expansion.

    Per Section 2.3 of [Genz2004bvnt]_, the bvnu equation (1) may be rewritten as

        `L(h, k, r) = L(h, k, s) - s/(2\pi) \int_{0}^{a} f(x) dx \
        f(x) = (1 - x^2){-1/2} e^{-0.5 ((h - sk)/ x)^2} e^{-shk/(1 + (1 - x^2)^{1/2})},`

    where `s = sign(r)` and `a = sqrt(1 - r^{2})`. The term `L(h, k, s)` is analytic.
    The second integral is approximated via Taylor expansion. See Sections 2.3 and
    2.4 of [Genz2004bvnt]_.
    """
    _0, _1, _ni2, _i2pi, _sq2pi = get_constants_like(
        values=(0, 1, -0.5, _inv_2pi, _sqrt_2pi), ref=r
    )

    x, w = leggauss(num_points, dtype=r.dtype, device=r.device)
    x = x + _1

    s = get_constants_like(2, r) * (r > _0).to(r) - _1  # sign of `r` where sign(0) := 1
    sk = s * k
    skh = sk * h
    comp_r2 = _1 - r.square()

    a = comp_r2.clip(min=0).sqrt()
    b = safe_sub(h, sk)
    b2 = b.square()
    c = get_constants_like(1 / 8, r) * (get_constants_like(4, r) - skh)
    d = get_constants_like(1 / 80, r) * (get_constants_like(12, r) - skh)

    # ---- Solve for `L(h, k, s)`
    int_from_0_to_s = case_dispatcher(
        out=torch.empty_like(r),
        cases=[(lambda: r > _0, lambda mask: Phi(-torch.maximum(h[mask], k[mask])))],
        default=lambda mask: (Phi(sk[mask]) - Phi(h[mask])).clip(min=_0),
    )

    # ---- Solve for `s/(2\pi) \int_{0}^{a} f(x) dx`
    # Analytic part
    _a0 = _ni2 * (safe_div(b2, comp_r2) + skh)
    _a1 = c * get_constants_like(1 / 3, r) * (_1 - d * b2)
    _a2 = _1 - b2 * _a1
    abs_b = b.abs()
    analytic_part = torch.subtract(  # analytic part of solution
        a * (_a2 + comp_r2 * _a1 + c * d * comp_r2.square()) * safe_exp(_a0),
        _sq2pi * Phi(safe_div(-abs_b, a)) * abs_b * _a2 * safe_exp(_ni2 * skh),
    )

    # Quadrature part
    _b2 = b2.unsqueeze(-1)
    _skh = skh.unsqueeze(-1)
    _q0 = get_constants_like(0.25, r) * comp_r2.unsqueeze(-1) * x.square()
    _q1 = (_1 - _q0).sqrt()
    _q2 = _ni2 * (_b2 / _q0 + _skh)

    _b2 = b2.unsqueeze(-1)
    _c = c.unsqueeze(-1)
    _d = d.unsqueeze(-1)
    vals = (_ni2 * (_b2 / _q0 + _skh)).exp() * torch.subtract(
        _1 + _c * _q0 * (_1 + get_constants_like(5, r) * _d * _q0),
        safe_exp(_ni2 * _q0 / (_1 + _q1).square() * _skh) / _q1,
    )
    mask = _q2 > get_constants_like(-100, r)
    if not mask.all():
        vals[~mask] = _0
    quadrature_part = _ni2 * a * (vals @ w)

    # Return `P(x > h, y > k)`
    int_from_0_to_a = _i2pi * s * (analytic_part + quadrature_part)
    return (int_from_0_to_s - int_from_0_to_a).clip(min=_0, max=_1)


def bvnmom(
    r: Tensor,
    xl: Tensor,
    yl: Tensor,
    xu: Tensor,
    yu: Tensor,
    p: Tensor | None = None,
) -> tuple[Tensor, Tensor]:
    r"""Computes the expected values of truncated, bivariate normal random variables.

    Let `x` and `y` be a pair of standard bivariate normal random variables having
    correlation `r`. This function computes `E([x,y] \| [xl,yl] < [x,y] < [xu,yu])`.

    Following [Muthen1990moments]_ equations (4) and (5), we have

        `E(x \| [xl, yl] < [x, y] < [xu, yu]) \
        = Z^{-1} \phi(xl) P(yl < y < yu \| x=xl) - \phi(xu) P(yl < y < yu \| x=xu),`

    where `Z = P([xl, yl] < [x, y] < [xu, yu])` and `\phi` is the standard normal PDF.

    Args:
        r: Tensor of correlation coefficients.
        xl: Tensor of lower bounds for `x`, same shape as `r`.
        xu: Tensor of upper bounds for `x`, same shape as `r`.
        yl: Tensor of lower bounds for `y`, same shape as `r`.
        yu: Tensor of upper bounds for `y`, same shape as `r`.
        p: Tensor of probabilities `P(xl < x < xu, yl < y < yu)`, same shape as `r`.

    Returns:
        `E(x \| [xl, yl] < [x, y] < [xu, yu])` and
        `E(y \| [xl, yl] < [x, y] < [xu, yu])`.
    """
    if not (r.shape == xl.shape == xu.shape == yl.shape == yu.shape):
        raise UnsupportedError("Arguments to `bvn` must have the same shape.")

    if p is None:
        p = bvn(r=r, xl=xl, xu=xu, yl=yl, yu=yu)

    corr = r[..., None, None]
    istd = (1 - corr.square()).rsqrt()
    lower = torch.stack([xl, yl], -1)
    upper = torch.stack([xu, yu], -1)
    bounds = torch.stack([lower, upper], -1)
    deltas = safe_mul(corr, bounds)

    # Compute densities and conditional probabilities
    density_at_bounds = phi(bounds)
    prob_given_bounds = Phi(
        safe_mul(istd, safe_sub(upper.flip(-1).unsqueeze(-1), deltas))
    ) - Phi(safe_mul(istd, safe_sub(lower.flip(-1).unsqueeze(-1), deltas)))

    # Evaluate Muthen's formula
    p_diffs = -(density_at_bounds * prob_given_bounds).diff().squeeze(-1)
    moments = (1 / p).unsqueeze(-1) * (p_diffs + r.unsqueeze(-1) * p_diffs.flip(-1))
    return moments.unbind(-1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Synthetic functions for multi-fidelity optimization benchmarks.

References:

.. [Chen2024]
    Chen, Y., et al. A latent variable approach for non-hierarchical
    multi-fidelity adaptive sampling. Computer Methods in Applied Mechanics
    and Engineering 421 (2024).
"""

from __future__ import annotations

import math

import torch
from botorch.test_functions.synthetic import SyntheticTestFunction
from torch import Tensor


class AugmentedBranin(SyntheticTestFunction):
    r"""Augmented Branin test function for multi-fidelity optimization.

    3-dimensional function with domain `[-5, 10] x [0, 15] * [0,1]`, where
    the last dimension of is the fidelity parameter:

        B(x) = (x_2 - (b - 0.1 * (1 - x_3))x_1^2 + c x_1 - r)^2 +
            10 (1-t) cos(x_1) + 10

    Here `b`, `c`, `r` and `t` are constants where `b = 5.1 / (4 * math.pi ** 2)`
    `c = 5 / math.pi`, `r = 6`, `t = 1 / (8 * math.pi)`.
    B has infinitely many minimizers with `x_1 = -pi, pi, 3pi`
    and `B_min = 0.397887`
    """

    dim = 3
    continuous_inds = list(range(dim))
    _bounds = [(-5.0, 10.0), (0.0, 15.0), (0.0, 1.0)]
    _optimal_value = 0.397887
    _optimizers = [  # this is a subset, ther are infinitely many optimizers
        (-math.pi, 12.275, 1),
        (math.pi, 1.3867356039019576, 0.1),
        (math.pi, 1.781519779945532, 0.5),
        (math.pi, 2.1763039559891064, 0.9),
    ]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        t1 = (
            X[..., 1]
            - (5.1 / (4 * math.pi**2) - 0.1 * (1 - X[..., 2])) * X[..., 0].pow(2)
            + 5 / math.pi * X[..., 0]
            - 6
        )
        t2 = 10 * (1 - 1 / (8 * math.pi)) * torch.cos(X[..., 0])
        return t1.pow(2) + t2 + 10


class AugmentedHartmann(SyntheticTestFunction):
    r"""Augmented Hartmann synthetic test function.

    7-dimensional function (typically evaluated on `[0, 1]^7`), where the last
    dimension is the fidelity parameter.

        H(x) = -(ALPHA_1 - 0.1 * (1-x_7)) * exp(- sum_{j=1}^6 A_1j (x_j - P_1j) ** 2) -
            sum_{i=2}^4 ALPHA_i exp( - sum_{j=1}^6 A_ij (x_j - P_ij) ** 2)

    H has a unique global minimizer
    `x = [0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573, 1.0]`

    with `H_min = -3.32237`
    """

    dim = 7
    continuous_inds = list(range(dim))
    _bounds = [(0.0, 1.0) for _ in range(dim)]
    _optimal_value = -3.32237
    _optimizers = [(0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573, 1.0)]
    _check_grad_at_opt = False

    def __init__(
        self,
        noise_std: float | None = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            dtype: The dtype that is used for the bounds of the function.
        """
        super().__init__(noise_std=noise_std, negate=negate, dtype=dtype)
        self.register_buffer("ALPHA", torch.tensor([1.0, 1.2, 3.0, 3.2]))
        A = [
            [10, 3, 17, 3.5, 1.7, 8],
            [0.05, 10, 17, 0.1, 8, 14],
            [3, 3.5, 1.7, 10, 17, 8],
            [17, 8, 0.05, 10, 0.1, 14],
        ]
        P = [
            [1312, 1696, 5569, 124, 8283, 5886],
            [2329, 4135, 8307, 3736, 1004, 9991],
            [2348, 1451, 3522, 2883, 3047, 6650],
            [4047, 8828, 8732, 5743, 1091, 381.0],
        ]
        self.register_buffer("A", torch.tensor(A))
        self.register_buffer("P", torch.tensor(P))

    def _evaluate_true(self, X: Tensor) -> Tensor:
        self.to(device=X.device, dtype=X.dtype)
        inner_sum = torch.sum(
            self.A * (X[..., :6].unsqueeze(-2) - 0.0001 * self.P).pow(2), dim=-1
        )
        alpha1 = self.ALPHA[0] - 0.1 * (1 - X[..., 6])
        H = (
            -(torch.sum(self.ALPHA[1:] * torch.exp(-inner_sum)[..., 1:], dim=-1))
            - alpha1 * torch.exp(-inner_sum)[..., 0]
        )
        return H


class AugmentedRosenbrock(SyntheticTestFunction):
    r"""Augmented Rosenbrock synthetic test function for multi-fidelity optimization.

    d-dimensional function (usually evaluated on `[-5, 10]^(d-2) * [0, 1]^2`),
    where the last two dimensions are the fidelity parameters:

        f(x) = sum_{i=1}^{d-1} (100 (x_{i+1} - x_i^2 + 0.1 * (1-x_{d-1}))^2 +
            (x_i - 1 + 0.1 * (1 - x_d)^2)^2)

    f has one minimizer for its global minimum at `z_1 = (1, 1, ..., 1)` with
    `f(z_i) = 0.0`.
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim: int = 3,
        noise_std: float | None = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension. Must be at least 3.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            dtype: The dtype that is used for the bounds of the function.
        """
        if dim < 3:
            raise ValueError(
                "AugmentedRosenbrock must be defined it at least 3 dimensions"
            )
        self.dim = dim
        self.continuous_inds = list(range(dim))
        self._bounds = [(-5.0, 10.0) for _ in range(self.dim - 2)] + [
            (0.0, 1.0) for _ in range(2)
        ]
        self._optimizers = [tuple(1.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        X_curr = X[..., :-3]
        X_next = X[..., 1:-2]
        t1 = 100 * (X_next - X_curr.pow(2) + 0.1 * (1 - X[..., -2:-1])).pow(2)
        t2 = (X_curr - 1 + 0.1 * (1 - X[..., -1:]).pow(2)).pow(2)
        return (t1 + t2).sum(dim=-1)


class WingWeightMultiFidelity(SyntheticTestFunction):
    """Wing Weight Design Problem from [Chen2024]_.

    Design variables (physical units):
       1. s_w        in [150,    200]  (wing area)
       2. w_fw       in [220,    300]  (fuel weight)
       3. A          in [6,       10]  (aspect ratio)
       4. Lambda_deg in [-10,     10]  (sweep angle, degrees)
       5. q          in [16,      45]  (dynamic pressure)
       6. lam        in [0.5,    1.0]  (taper ratio)
       7. t_c        in [0.08,  0.18]   (thickness-to-chord)
       8. N_z        in [2.5,    6.0]  (ultimate load factor)
       9. w_dg       in [1700,  2500]  (design gross weight)
       10. w_pp      in [0.025, 0.08]  (weight per unit area)

    Fidelity parameter (stored as the 11th input):
      0: High fidelity (HF)
      1: Low fidelity 1 (LF1)
      2: Low fidelity 2 (LF2)
      3: Low fidelity 3 (LF3)

    LF models use slightly altered exponents and additive biases.
    """

    dim = 11
    continuous_inds = list(range(dim))
    _num_fidelities = 1
    _bounds = [
        (150.0, 200.0),  # s_w
        (220.0, 300.0),  # w_fw
        (6.0, 10.0),  # A
        (-10.0, 10.0),  # Lambda_deg
        (16.0, 45.0),  # q
        (0.5, 1.0),  # lam
        (0.08, 0.18),  # t_c
        (2.5, 6.0),  # N_z
        (1700.0, 2500.0),  # w_dg
        (0.025, 0.08),  # w_pp
        (0, 3),  # fidelity
    ]
    fidelities = [0, 1, 2, 3]
    _optimal_value = 123.25

    def _evaluate_true(self, X: torch.Tensor) -> Tensor:
        s_w, w_fw, A, Lambda_deg, q, lam, t_c, N_z, w_dg, w_pp, fidelity = X.unbind(
            dim=-1
        )
        Lambda_rad = Lambda_deg * math.pi / 180.0
        cos_val = torch.cos(Lambda_rad)
        y = torch.zeros_like(s_w)
        shared_multiplier = (
            0.036
            * q**0.006
            * lam**0.04
            * (A / (cos_val**2)) ** 0.6
            * (100.0 * t_c / cos_val) ** (-0.3)
            * (N_z * w_dg) ** 0.49
            * (w_fw**0.0035)
        )
        # High fidelity (fidelity == 0)
        mask = fidelity == 0
        if mask.any():
            hf = s_w**0.758 * shared_multiplier + s_w * w_pp
            y[mask] = hf[mask]
        # Low fidelity 1 (fidelity == 1)
        mask = fidelity == 1
        if mask.any():
            lf1 = s_w**0.758 * shared_multiplier + w_pp
            y[mask] = lf1[mask]
        # Low fidelity 2 (fidelity == 2)
        mask = fidelity == 2
        if mask.any():
            lf2 = s_w**0.8 * shared_multiplier + w_pp
            y[mask] = lf2[mask]
        # Low fidelity 3 (fidelity == 3)
        mask = fidelity == 3
        if mask.any():
            lf3 = s_w**0.9 * shared_multiplier
            y[mask] = lf3[mask]
        return y

    def cost(self, X: torch.Tensor) -> Tensor:
        fidelity = X[..., 10]
        c = torch.zeros_like(fidelity)
        c[fidelity == 0] = 1000.0
        c[fidelity == 1] = 100.0
        c[fidelity == 2] = 10.0
        c[fidelity == 3] = 1.0
        return c


class BoreholeMultiFidelity(SyntheticTestFunction):
    """Borehole Problem from [Chen2024]_.

    This problem models water flow through a borehole with 8 design variables:
      1. r_w   in [0.05,   0.15]   (borehole radius)
      2. r     in [100,    50000]  (radius of influence)
      3. T_u   in [63070,  115600] (transmissivity of upper aquifer)
      4. T_l   in [63.1,   116]    (transmissivity of lower aquifer)
      5. H_u   in [990,    1110]   (potentiometric head of upper aquifer)
      6. H_l   in [700,    820]    (potentiometric head of lower aquifer)
      7. L     in [1120,   1680]   (length of borehole)
      8. K_w   in [9855,   12045]  (hydraulic conductivity)

    The fidelity index (9th input) is categorical:
      0: High fidelity (HF)
      1: Low fidelity 1 (LF1)
      2: Low fidelity 2 (LF2)
      3: Low fidelity 3 (LF3)
      4: Low fidelity 4 (LF4)


    The low-fidelity models modify exponents and add a bias.
    """

    dim = 9
    continuous_inds = list(range(dim))
    _num_fidelities = 1
    _bounds = [
        (0.05, 0.15),  # r_w
        (100.0, 10000.0),  # r
        (100.0, 1000.0),  # T_u
        (10.0, 500.0),  # T_l
        (990.0, 1110.0),  # H_u
        (700.0, 820.0),  # H_l
        (1000.0, 2000.0),  # L
        (6000.0, 12000.0),  # K_w
        (0, 4),  # fidelity
    ]
    fidelities = [0, 1, 2, 3, 4]
    _optimal_value = 3.98

    def _evaluate_true(self, X: torch.Tensor) -> torch.Tensor:
        r_w, r, T_u, T_l, H_u, H_l, L, K_w, fidelity = X.unbind(dim=-1)
        LTu = L * T_u
        two_pi_T_u = 2.0 * math.pi * T_u
        log_term = torch.log(r / r_w)
        denom = log_term * (r_w**2) * K_w
        numer = two_pi_T_u * (H_u - H_l)
        T_u_over_T_l = T_u / T_l
        y = torch.zeros_like(r_w)
        # HF (fidelity 0)
        mask = fidelity == 0
        if mask.any():
            hf_denom = log_term * (1.0 + (2.0 * LTu) / denom + T_u_over_T_l)
            hf = numer / hf_denom
            y[mask] = hf[mask]

        # LF1 (fidelity 1): add bias.
        mask = fidelity == 1
        if mask.any():
            lf1_numer = two_pi_T_u * (H_u - 0.8 * H_l)
            lf1_denom = log_term * (1.0 + LTu / denom + T_u_over_T_l)
            lf1 = lf1_numer / lf1_denom
            y[mask] = lf1[mask]

        # LF2 (fidelity 2): modify the exponent on log_term and add bias.
        mask = fidelity == 2
        if mask.any():
            lf2_denom = log_term * (1.0 + (8 * LTu) / denom + 0.75 * T_u_over_T_l)
            lf2 = numer / lf2_denom
            y[mask] = lf2[mask]

        # LF3 (fidelity 3): modify r_w exponent slightly.
        mask = fidelity == 3
        if mask.any():
            lf3_log_term = torch.log(4 * r / r_w)
            lf3_numer = two_pi_T_u * (1.09 * H_u - H_l)
            lf3_denom = lf3_log_term * (1.0 + (3 * LTu) / denom + T_u_over_T_l)
            lf3 = lf3_numer / lf3_denom
            y[mask] = lf3[mask]
        # LF4 (fidelity 4): further bias.
        mask = fidelity == 4
        if mask.any():
            lf4_log_term = torch.log(2 * r / r_w)
            lf4_numer = two_pi_T_u * (1.05 * H_u - H_l)
            lf4_denom = lf4_log_term * (1.0 + (3 * LTu) / denom + T_u_over_T_l)
            lf4 = lf4_numer / lf4_denom
            y[mask] = lf4[mask]

        return y

    def cost(self, X: torch.Tensor) -> torch.Tensor:
        fidelity = X[..., 8]
        c = torch.zeros_like(fidelity)
        c[fidelity == 0] = 1000.0
        c[fidelity == 1] = 100.0
        c[fidelity == 2] = 10.0
        c[fidelity == 3] = 100.0
        c[fidelity == 4] = 10.0
        return c



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import torch
from botorch.exceptions.errors import BotorchError
from botorch.posteriors.base_samples import _reshape_base_samples_non_interleaved
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions.multitask_multivariate_normal import (
    MultitaskMultivariateNormal,
)
from linear_operator.operators import BlockDiagLinearOperator, LinearOperator

from linear_operator.utils.cholesky import psd_safe_cholesky
from linear_operator.utils.errors import NanError
from torch import Tensor


def extract_batch_covar(mt_mvn: MultitaskMultivariateNormal) -> LinearOperator:
    r"""Extract a batched independent covariance matrix from an MTMVN.

    Args:
        mt_mvn: A multi-task multivariate normal with a block diagonal
            covariance matrix.

    Returns:
        A lazy covariance matrix consisting of a batch of the blocks of
            the diagonal of the MultitaskMultivariateNormal.

    """
    lazy_covar = mt_mvn.lazy_covariance_matrix
    if not isinstance(lazy_covar, BlockDiagLinearOperator):
        raise BotorchError(
            f"Expected BlockDiagLinearOperator, but got {type(lazy_covar)}."
        )
    return lazy_covar.base_linear_op


def _reshape_base_samples(
    base_samples: Tensor, sample_shape: torch.Size, posterior: GPyTorchPosterior
) -> Tensor:
    r"""Manipulate shape of base_samples to match `MultivariateNormal.rsample`.

    This ensure that base_samples are used in the same way as in
    gpytorch.distributions.MultivariateNormal. For CBD, it is important to ensure
    that the same base samples are used for the in-sample points here and in the
    cached box decompositions.

    Args:
        base_samples: The base samples.
        sample_shape: The sample shape.
        posterior: The joint posterior is over (X_baseline, X).

    Returns:
        Reshaped and expanded base samples.
    """
    mvn = posterior.distribution
    loc = mvn.loc
    peshape = posterior._extended_shape()
    base_samples = base_samples.view(
        sample_shape + torch.Size([1] * (loc.ndim - 1)) + peshape[-2:]
    ).expand(sample_shape + loc.shape[:-1] + peshape[-2:])
    if posterior._is_mt:
        base_samples = _reshape_base_samples_non_interleaved(
            mvn=posterior.distribution,
            base_samples=base_samples,
            sample_shape=sample_shape,
        )
    base_samples = base_samples.reshape(
        -1, *loc.shape[:-1], mvn.lazy_covariance_matrix.shape[-1]
    )
    base_samples = base_samples.permute(*range(1, loc.dim() + 1), 0)
    return base_samples.reshape(
        *peshape[:-2],
        peshape[-1],
        peshape[-2],
        *sample_shape,
    )


def sample_cached_cholesky(
    posterior: GPyTorchPosterior,
    baseline_L: Tensor,
    q: int,
    base_samples: Tensor,
    sample_shape: torch.Size,
    max_tries: int = 6,
) -> Tensor:
    r"""Get posterior samples at the `q` new points from the joint multi-output
    posterior.

    Args:
        posterior: The joint posterior is over (X_baseline, X).
        baseline_L: The baseline lower triangular cholesky factor.
        q: The number of new points in X.
        base_samples: The base samples.
        sample_shape: The sample shape.
        max_tries: The number of tries for computing the Cholesky
            decomposition with increasing jitter.


    Returns:
        A `sample_shape x batch_shape x q x m`-dim tensor of posterior
            samples at the new points.
    """
    # compute bottom left covariance block
    mvn = posterior.distribution
    lazy_covar = (
        extract_batch_covar(mt_mvn=mvn)
        if isinstance(mvn, MultitaskMultivariateNormal)
        else mvn.lazy_covariance_matrix
    )
    # Get the `q` new rows of the batched covariance matrix
    bottom_rows = lazy_covar[..., -q:, :].to_dense()
    # The covariance in block form is:
    # [K(X_baseline, X_baseline), K(X_baseline, X)]
    # [K(X, X_baseline), K(X, X)]
    # bl := K(X, X_baseline)
    # br := K(X, X)
    # Get bottom right block of new covariance
    bl, br = bottom_rows.split([bottom_rows.shape[-1] - q, q], dim=-1)
    # Solve Ax = b
    # where A = K(X_baseline, X_baseline) and b = K(X, X_baseline)^T
    # and bl_chol := x^T
    # bl_chol is the new `(batch_shape) x q x n`-dim bottom left block
    # of the cholesky decomposition
    bl_chol = torch.linalg.solve_triangular(
        baseline_L, bl.transpose(-2, -1), upper=False
    ).transpose(-2, -1)

    # Compute the new bottom right block of the Cholesky
    # decomposition via:
    # Cholesky(K(X, X) - bl_chol @ bl_chol^T)
    br_to_chol = br - bl_chol @ bl_chol.transpose(-2, -1)
    # TODO: technically we should make sure that we add a
    # consistent nugget to the cached covariance and the new block
    br_chol = psd_safe_cholesky(br_to_chol, max_tries=max_tries)
    # Create a `(batch_shape) x q x (n+q)`-dim tensor containing the
    # `q` new bottom rows of the Cholesky decomposition
    new_Lq = torch.cat([bl_chol, br_chol], dim=-1)
    mean = posterior.distribution.mean
    base_samples = _reshape_base_samples(
        base_samples=base_samples,
        sample_shape=sample_shape,
        posterior=posterior,
    )
    if not isinstance(posterior.distribution, MultitaskMultivariateNormal):
        # add output dim
        mean = mean.unsqueeze(-1)
        # add batch dim corresponding to output dim
        new_Lq = new_Lq.unsqueeze(-3)
    new_mean = mean[..., -q:, :]
    res = (
        new_Lq.matmul(base_samples)
        .add(new_mean.transpose(-1, -2).unsqueeze(-1))
        .permute(-1, *range(posterior.distribution.loc.dim() - 1), -2, -3)
        .contiguous()
    )
    contains_nans = torch.isnan(res).any()
    contains_infs = torch.isinf(res).any()
    if contains_nans or contains_infs:
        suffix_args = []
        if contains_nans:
            suffix_args.append("nans")
        if contains_infs:
            suffix_args.append("infs")
        suffix = " and ".join(suffix_args)
        raise NanError(f"Samples contain {suffix}.")
    return res



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Acquisition functions for max-value entropy search for multi-objective
Bayesian optimization (MESMO).
"""

from __future__ import annotations

from math import pi

import torch
from botorch.acquisition.max_value_entropy_search import qMaxValueEntropy
from botorch.acquisition.multi_objective.base import MultiObjectiveMCAcquisitionFunction
from botorch.acquisition.multi_objective.joint_entropy_search import (
    LowerBoundMultiObjectiveEntropySearch,
)
from botorch.models.model import Model
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.utils.transforms import (
    average_over_ensemble_models,
    concatenate_pending_points,
    t_batch_mode_transform,
)
from torch import Tensor


# Can be removed in version 0.15.0, or potentially sooner because the code has
# already been raising deprecation warnings for a long time
class qMultiObjectiveMaxValueEntropy(
    qMaxValueEntropy, MultiObjectiveMCAcquisitionFunction
):
    r"""The acquisition function for MESMO.

    This is no longer available. We recommend
    `qLowerBoundMultiObjectiveMaxValueEntropySearch` as a replacement.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Multi-objective max-value entropy search acquisition function."""
        raise NotImplementedError(
            "qMultiObjectiveMaxValueEntropy is no longer available. We suggest "
            "qLowerBoundMultiObjectiveMaxValueEntropySearch as a replacement."
        )


class qLowerBoundMultiObjectiveMaxValueEntropySearch(
    LowerBoundMultiObjectiveEntropySearch
):
    r"""The acquisition function for the multi-objective Max-value Entropy Search,
    where the batches `q > 1` are supported through the lower bound formulation.

    This acquisition function computes the mutual information between the observation
    at a candidate point `X` and the Pareto optimal outputs.

    See [Tu2022]_ for a discussion on the estimation procedure.

    NOTES:
    (i) The estimated acquisition value could be negative.

    (ii) The lower bound batch acquisition function might not be monotone in the
    sense that adding more elements to the batch does not necessarily increase the
    acquisition value. Specifically, the acquisition value can become smaller when
    more inputs are added.
    """

    def __init__(
        self,
        model: Model,
        hypercell_bounds: Tensor,
        X_pending: Tensor | None = None,
        estimation_type: str = "LB",
        num_samples: int = 64,
    ) -> None:
        r"""Lower bound multi-objective max-value entropy search acquisition function.

        Args:
            model: A fitted batch model with 'M' number of outputs.
            hypercell_bounds:  A `num_pareto_samples x 2 x J x M`-dim Tensor
                containing the hyper-rectangle bounds for integration, where `J` is
                the number of hyper-rectangles. In the unconstrained case, this gives
                the partition of the dominated space. In the constrained case, this
                gives the partition of the feasible dominated space union the
                infeasible space.
            X_pending: A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation, but have not yet been evaluated.
            estimation_type: A string to determine which entropy estimate is
                computed: "0", "LB", "LB2", or "MC".
            num_samples: The number of Monte Carlo samples for the Monte Carlo
                estimate.
        """
        super().__init__(
            model=model,
            pareto_sets=None,
            pareto_fronts=None,
            hypercell_bounds=hypercell_bounds,
            X_pending=X_pending,
            estimation_type=estimation_type,
            num_samples=num_samples,
        )

    def _compute_posterior_statistics(
        self, X: Tensor
    ) -> dict[str, GPyTorchPosterior | Tensor]:
        r"""Compute the posterior statistics.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of inputs.

        Returns:
            A dictionary containing the posterior variables used to estimate the
            entropy.

            - "initial_entropy": A `batch_shape`-dim Tensor containing the entropy of
                the Gaussian random variable `p(Y| X, D_n)`.
            - "posterior_mean": A `batch_shape x num_pareto_samples x q x 1 x M`-dim
                Tensor containing the posterior mean at the input `X`.
            - "posterior_variance": A `batch_shape x num_pareto_samples x q x 1 x
                M`-dim Tensor containing the posterior variance at the input `X`
                excluding the observation noise.
            - "observation_noise": A `batch_shape x num_pareto_samples x q x 1 x M`
                -dim Tensor containing the observation noise at the input `X`.
            - "posterior_with_noise": The posterior distribution at `X` which
                includes the observation noise. This is used to compute the marginal
                log-probabilities with respect to `p(y| x, D_n)` for `x` in `X`.
        """
        tkwargs = {"dtype": X.dtype, "device": X.device}
        CLAMP_LB = torch.finfo(tkwargs["dtype"]).eps

        # Compute the initial entropy term depending on `X`.
        # TODO: Below we compute posterior_plus_noise twice:
        #  (1) Firstly, we compute p(Y| X, D_n) when computing the initial entropy
        #  (2) Secondly, we compute p(y| x, D_n) for x in X in order to compute
        #  log(p(y|x, D_n)) for x in X in the Monte Carlo estimate..
        #  This could be simplified if we could evaluate log(p(y|x, D_n)) using the
        #  the posterior p(Y| X, D_n)
        posterior_plus_noise = self.initial_model.posterior(X, observation_noise=True)

        # Additional constant term.
        add_term = (
            0.5
            * self.model.num_outputs
            * (1 + torch.log(2 * pi * torch.ones(1, **tkwargs)))
        )
        # The variance initially has shape `batch_shape x (q*M) x (q*M)`
        # prior_entropy has shape `batch_shape x num_fantasies`
        initial_entropy = add_term + 0.5 * torch.logdet(
            posterior_plus_noise.mvn.covariance_matrix
        )
        posterior_statistics = {"initial_entropy": initial_entropy}

        # Compute the posterior entropy term.
        posterior_plus_noise = self.model.posterior(
            X.unsqueeze(-2), observation_noise=True
        )

        # `batch_shape x q x 1 x M`
        mean = posterior_plus_noise.mean
        var_plus_noise = posterior_plus_noise.variance.clamp_min(CLAMP_LB)
        # Expand shapes to `batch_shape x num_pareto_samples x q x 1 x M`
        new_shape = (
            mean.shape[:-3] + torch.Size([self.num_pareto_samples]) + mean.shape[-3:]
        )
        mean = mean.unsqueeze(-4).expand(new_shape)
        var_plus_noise = var_plus_noise.unsqueeze(-4).expand(new_shape)

        # TODO: This computes the observation noise via a second evaluation of the
        #   posterior. This step could be done better.
        posterior = self.model.posterior(X.unsqueeze(-2), observation_noise=False)
        var = posterior.variance.clamp_min(CLAMP_LB)
        var = var.unsqueeze(-4).expand(new_shape)
        obs_noise = var_plus_noise - var

        posterior_statistics["posterior_mean"] = mean
        posterior_statistics["posterior_variance"] = var
        posterior_statistics["observation_noise"] = obs_noise
        posterior_statistics["posterior_with_noise"] = posterior_plus_noise

        return posterior_statistics

    def _compute_monte_carlo_variables(
        self, posterior: GPyTorchPosterior
    ) -> tuple[Tensor, Tensor]:
        r"""Compute the samples and log-probability associated with a posterior
        distribution.

        Args:
            posterior: The posterior distribution, which includes the observation
                noise.

        Returns:
            A two-element tuple containing

            - samples: A `num_mc_samples x batch_shape x num_pareto_samples x q x 1
                x M`-dim Tensor containing the Monte Carlo samples.
            - samples_log_prob: A `num_mc_samples x batch_shape x num_pareto_samples
                x q`-dim Tensor containing the log-probabilities of the Monte Carlo
                samples.
        """

        # `num_mc_samples x batch_shape x q x 1 x M`
        samples = self.get_posterior_samples(posterior)

        # `num_mc_samples x batch_shape x q`
        if self.model.num_outputs == 1:
            samples_log_prob = posterior.mvn.log_prob(samples.squeeze(-1))
        else:
            samples_log_prob = posterior.mvn.log_prob(samples)

        # Expand shape to `num_mc_samples x batch_shape x num_pareto_samples x
        # q x 1 x M`
        new_shape = (
            samples.shape[:-3]
            + torch.Size([self.num_pareto_samples])
            + samples.shape[-3:]
        )
        samples = samples.unsqueeze(-4).expand(new_shape)

        # Expand shape to `num_mc_samples x batch_shape x num_pareto_samples x q`
        new_shape = (
            samples_log_prob.shape[:-1]
            + torch.Size([self.num_pareto_samples])
            + samples_log_prob.shape[-1:]
        )
        samples_log_prob = samples_log_prob.unsqueeze(-2).expand(new_shape)

        return samples, samples_log_prob

    @concatenate_pending_points
    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluates qLowerBoundMultiObjectiveMaxValueEntropySearch at the design
        points `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of `batch_shape` t-batches with `q`
            `d`-dim design points each.

        Returns:
            A `batch_shape`-dim Tensor of acquisition values at the given design
            points `X`.
        """
        return self._compute_lower_bound_information_gain(X)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Candidate generation utilities.
"""

from __future__ import annotations

import time
import warnings
from collections.abc import Callable
from functools import partial
from typing import Any, Mapping, NoReturn

import numpy as np
import numpy.typing as npt
import torch
from botorch.acquisition import AcquisitionFunction
from botorch.exceptions.errors import OptimizationGradientError, UnsupportedError
from botorch.exceptions.warnings import OptimizationWarning
from botorch.generation.utils import _remove_fixed_features_from_optimization
from botorch.logging import logger
from botorch.optim.parameter_constraints import (
    _arrayify,
    make_scipy_bounds,
    make_scipy_linear_constraints,
    make_scipy_nonlinear_inequality_constraints,
)
from botorch.optim.stopping import ExpMAStoppingCriterion
from botorch.optim.utils import (
    check_scipy_version_at_least,
    columnwise_clamp,
    fix_features,
    minimize_with_timeout,
)
from scipy.optimize import OptimizeResult
from threadpoolctl import threadpool_limits
from torch import Tensor
from torch.optim import Optimizer

if check_scipy_version_at_least(minor=13) and not check_scipy_version_at_least(
    minor=17
):
    # We only import the batched lbfgs_b code here, as it might otherwise
    # lead to import errors, if the wrong scipy version is used
    from botorch.optim.batched_lbfgs_b import (
        fmin_l_bfgs_b_batched,
        translate_bounds_for_lbfgsb,
    )

TGenCandidates = Callable[[Tensor, AcquisitionFunction, Any], tuple[Tensor, Tensor]]


def gen_candidates_scipy(
    initial_conditions: Tensor,
    acquisition_function: AcquisitionFunction,
    lower_bounds: float | Tensor | None = None,
    upper_bounds: float | Tensor | None = None,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    nonlinear_inequality_constraints: list[tuple[Callable, bool]] | None = None,
    options: dict[str, Any] | None = None,
    fixed_features: Mapping[int, float | Tensor] | None = None,
    timeout_sec: float | None = None,
    use_parallel_mode: bool | None = None,
) -> tuple[Tensor, Tensor]:
    r"""Generate a set of candidates using `scipy.optimize.minimize`.

    Optimizes an acquisition function starting from a set of initial candidates
    using `scipy.optimize.minimize` via a numpy converter.
    We use SLSQP, if constraints are present, and LBFGS-B otherwise.
    As `scipy.optimize.minimize` does not support optimizating a batch of problems, we
    treat optimizing a set of candidates as a single optimization problem by
    summing together their acquisition values.

    Args:
        initial_conditions: Starting points for optimization, with shape
            (b) x q x d.
        acquisition_function: Acquisition function to be used.
        lower_bounds: Minimum values for each column of initial_conditions.
        upper_bounds: Maximum values for each column of initial_conditions.
        inequality constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) >= rhs`.
        equality constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) = rhs`.
        nonlinear_inequality_constraints: A list of tuples representing the nonlinear
            inequality constraints. The first element in the tuple is a callable
            representing a constraint of the form `callable(x) >= 0`. In case of an
            intra-point constraint, `callable()`takes in an one-dimensional tensor of
            shape `d` and returns a scalar. In case of an inter-point constraint,
            `callable()` takes a two dimensional tensor of shape `q x d` and again
            returns a scalar. The second element is a boolean, indicating if it is an
            intra-point or inter-point constraint (`True` for intra-point. `False` for
            inter-point). For more information on intra-point vs inter-point
            constraints, see the docstring of the `inequality_constraints` argument to
            `optimize_acqf()`. The constraints will later be passed to the scipy
            solver.
        options: Options used to control the optimization including "method"
            and "maxiter". Select method for `scipy.optimize.minimize` using the
            "method" key. By default uses L-BFGS-B for box-constrained problems
            and SLSQP if inequality or equality constraints are present. If
            `with_grad=False`, then we use a two-point finite difference estimate
            of the gradient.
        fixed_features: Mapping[int, float | Tensor] | None,
            all generated candidates will have features fixed to these values.
            If passing tensors as values, they should have either shape `b` or
            `b x q` to fix the same feature to different values in the batch.
            Assumes values to be compatible with lower_bounds and upper_bounds!
        timeout_sec: Timeout (in seconds) for `scipy.optimize.minimize` routine -
            if provided, optimization will stop after this many seconds and return
            the best solution found so far.
        use_parallel_mode: If None uses the parallel implementation of l-bfgs-b,
            if possible.
            If True, forces the use of the parallel implementation and fails if not
            possible.
            If using parallel mode, each item in the batch dimension is treated as a
            separate optimization problem, we enforce the shape of the initial
            conditions to be `b x q x d` or `q x d`, and we assume the
            `acquisition_function` does not treat elements differently in the batch
            dimension (it is simply a batched function).
            If False, forces the use of the serial implementation through
            `scipy.optimize.minimize`.

    Returns:
        2-element tuple containing

        - The set of generated candidates.
        - The acquisition value for each t-batch.

    Example:
        >>> qEI = qExpectedImprovement(model, best_f=0.2)
        >>> bounds = torch.tensor([[0., 0.], [1., 2.]])
        >>> Xinit = gen_batch_initial_conditions(
        >>>     qEI, bounds, q=3, num_restarts=25, raw_samples=500
        >>> )
        >>> batch_candidates, batch_acq_values = gen_candidates_scipy(
                initial_conditions=Xinit,
                acquisition_function=qEI,
                lower_bounds=bounds[0],
                upper_bounds=bounds[1],
            )
    """
    if use_parallel_mode:
        if initial_conditions.ndim not in (2, 3):
            raise UnsupportedError(
                "Initial conditions must have either 2 or 3 dimensions. "
                f"Received shape {initial_conditions.shape}"
            )
    options = options or {}
    options = {**options, "maxiter": options.get("maxiter", 2000)}

    original_initial_conditions_shape = initial_conditions.shape

    if initial_conditions.ndim == 2:
        initial_conditions = initial_conditions.unsqueeze(0)

    initial_conditions_all_features = initial_conditions
    if fixed_features:
        initial_conditions = initial_conditions[
            ...,
            [i for i in range(initial_conditions.shape[-1]) if i not in fixed_features],
        ]
        if isinstance(lower_bounds, Tensor):
            lower_bounds = lower_bounds[
                [i for i in range(len(lower_bounds)) if i not in fixed_features]
            ]
        if isinstance(upper_bounds, Tensor):
            upper_bounds = upper_bounds[
                [i for i in range(len(upper_bounds)) if i not in fixed_features]
            ]

    clamped_candidates = columnwise_clamp(
        X=initial_conditions,
        lower=lower_bounds,
        upper=upper_bounds,
        raise_on_violation=True,
    )

    def f(x):
        return -acquisition_function(x)

    is_constrained = (
        nonlinear_inequality_constraints
        or equality_constraints
        or inequality_constraints
    )
    method = options.get("method", "SLSQP" if is_constrained else "L-BFGS-B")
    with_grad = options.get("with_grad", True)
    minimize_options = {
        k: v
        for k, v in options.items()
        if k
        not in [
            "method",
            "callback",
            "with_grad",
            "max_optimization_problem_aggregation_size",
        ]
    }

    why_not_fast_path = get_reasons_against_fast_path(
        method=method,
        with_grad=with_grad,
        minimize_options=minimize_options,
        timeout_sec=timeout_sec,
    )

    f_np_wrapper = _get_f_np_wrapper(
        clamped_candidates.shape,
        initial_conditions.device,
        initial_conditions.dtype,
        with_grad,
    )

    if not why_not_fast_path and use_parallel_mode is not False:
        if is_constrained:
            raise RuntimeWarning("Method L-BFGS-B cannot handle constraints.")

        batched_x0 = _arrayify(clamped_candidates).reshape(len(clamped_candidates), -1)

        l_bfgs_b_bounds = translate_bounds_for_lbfgsb(
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            num_features=clamped_candidates.shape[-1],
            q=clamped_candidates.shape[1],
        )

        with threadpool_limits(limits=1, user_api="blas"):
            xs, fs, results = fmin_l_bfgs_b_batched(
                func=partial(f_np_wrapper, f=f, fixed_features=fixed_features),
                # args is not necessary, done via the partial instead
                x0=batched_x0,
                # method=method, # method is not necessary as it is only l-bfgs-b
                # jac=with_grad, this is assumed to be true
                bounds=l_bfgs_b_bounds,
                # constraints=constraints,
                callback=options.get("callback", None),
                pass_batch_indices=True,
                **minimize_options,
            )
        for res in results:
            _process_scipy_result(res=res, options=options)

    else:
        # In this case we optimize multiple initial conditions in a single
        # problem, up to max_optimization_problem_aggregation_size at a time.
        max_optimization_problem_aggregation_size = options.get(
            "max_optimization_problem_aggregation_size", len(clamped_candidates)
        )

        if use_parallel_mode is not False:
            msg = (
                "Not using the parallel implementation of l-bfgs-b, as: "
                + ", and ".join(why_not_fast_path)
            )
            if use_parallel_mode:
                raise NotImplementedError(msg)
            else:
                logger.debug(msg)

        if (
            fixed_features
            and any(
                torch.is_tensor(ff) and ff.ndim > 0 for ff in fixed_features.values()
            )
            and max_optimization_problem_aggregation_size != 1
        ):
            raise UnsupportedError(
                "Batch shaped fixed features are not "
                "supported, when optimizing more than one optimization "
                "problem at a time."
            )

        all_xs = []
        split_candidates = clamped_candidates.split(
            max_optimization_problem_aggregation_size
        )
        for i, candidates_ in enumerate(split_candidates):
            if fixed_features:
                fixed_features_ = {
                    k: ff[i : i + 1].item()
                    # from the test above, we know that we only treat one candidate
                    # at a time thus we can use index i
                    if torch.is_tensor(ff) and ff.ndim > 0
                    else ff
                    for k, ff in fixed_features.items()
                }
            else:
                fixed_features_ = None

            _no_fixed_features = _remove_fixed_features_from_optimization(
                fixed_features=fixed_features_,
                acquisition_function=acquisition_function,
                initial_conditions=None,
                d=initial_conditions_all_features.shape[-1],
                lower_bounds=None,
                upper_bounds=None,
                inequality_constraints=inequality_constraints,
                equality_constraints=equality_constraints,
                nonlinear_inequality_constraints=nonlinear_inequality_constraints,
            )
            bounds = make_scipy_bounds(
                X=candidates_,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
            )

            f_np_wrapper_ = partial(
                f_np_wrapper,
                fixed_features=fixed_features_,
            )

            x0 = candidates_.flatten()

            constraints = make_scipy_linear_constraints(
                shapeX=candidates_.shape,
                inequality_constraints=_no_fixed_features.inequality_constraints,
                equality_constraints=_no_fixed_features.equality_constraints,
            )
            if _no_fixed_features.nonlinear_inequality_constraints:
                # Make sure `batch_limit` is 1 for now.
                if not (len(candidates_.shape) == 3 and candidates_.shape[0] == 1):
                    raise ValueError(
                        "`batch_limit` must be 1 when non-linear inequality "
                        "constraints are given."
                    )
                nl_ineq_constraints = (
                    _no_fixed_features.nonlinear_inequality_constraints
                )
                constraints += make_scipy_nonlinear_inequality_constraints(
                    nonlinear_inequality_constraints=nl_ineq_constraints,
                    f_np_wrapper=f_np_wrapper_,
                    x0=x0,
                    shapeX=candidates_.shape,
                )

            x0 = _arrayify(x0)

            res = minimize_with_timeout(
                fun=f_np_wrapper_,
                args=(f,),
                x0=x0,
                method=method,
                jac=with_grad,
                bounds=bounds,
                constraints=constraints,
                callback=options.get("callback", None),
                options=minimize_options,
                timeout_sec=timeout_sec / len(split_candidates)
                if timeout_sec is not None
                else None,
            )
            _process_scipy_result(res=res, options=options)
            xs = res.x.reshape(candidates_.shape)
            all_xs.append(xs)
        xs = np.concatenate(all_xs)

    candidates = torch.from_numpy(xs).view_as(clamped_candidates).to(initial_conditions)

    clamped_candidates = columnwise_clamp(
        X=candidates, lower=lower_bounds, upper=upper_bounds, raise_on_violation=True
    )

    clamped_candidates = fix_features(
        X=clamped_candidates,
        fixed_features=fixed_features,
        replace_current_value=False,
    )
    clamped_candidates = clamped_candidates.reshape(original_initial_conditions_shape)

    with torch.no_grad():
        batch_acquisition = acquisition_function(clamped_candidates)

    return clamped_candidates, batch_acquisition


def _get_f_np_wrapper(shapeX, device, dtype, with_grad):
    if with_grad:

        def f_np_wrapper(
            x: npt.NDArray,
            f: Callable,
            fixed_features: Mapping[int, float | Tensor] | None,
            batch_indices: list[int] | None = None,
        ) -> tuple[float | np.NDArray, np.NDArray]:
            """Given a torch callable, compute value + grad given a numpy array."""
            if np.isnan(x).any():
                raise RuntimeError(
                    f"{np.isnan(x).sum()} elements of the {x.size} element array "
                    f"`x` are NaN."
                )
            X = (
                torch.from_numpy(x)
                .to(device=device, dtype=dtype)
                # We reshape in this way, as in parallel mode the batch dimension might
                # change during optimization: some examples might finish earlier than
                # others.
                .view(-1, *shapeX[1:])
                .contiguous()
                .requires_grad_(True)
            )
            if fixed_features is not None:
                if batch_indices is not None:
                    this_fixed_features = {
                        k: ff[batch_indices]
                        if torch.is_tensor(ff) and ff.ndim > 0
                        else ff
                        for k, ff in fixed_features.items()
                    }
                else:
                    this_fixed_features = fixed_features
            else:
                this_fixed_features = None

            X_fix = fix_features(
                X, fixed_features=this_fixed_features, replace_current_value=False
            )
            # we compute the loss on the whole batch, under the assumption that f
            # treats multiple inputs in the 0th dimension as independent
            # inputs in a batch
            losses = f(X_fix)
            loss = losses.sum()
            # compute gradient w.r.t. the inputs (does not accumulate in leaves)
            gradf = _arrayify(torch.autograd.grad(loss, X)[0].contiguous().view(-1))
            gradf = gradf.reshape(*x.shape)
            if np.isnan(gradf).any():
                msg = (
                    f"{np.isnan(gradf).sum()} elements of the {x.size} element "
                    "gradient array `gradf` are NaN. "
                    "This often indicates numerical issues."
                )
                if x.dtype != torch.double:
                    msg += " Consider using `dtype=torch.double`."
                raise OptimizationGradientError(msg, current_x=x)
            fval = (
                losses.detach().view(-1).cpu().numpy()
                if batch_indices is not None
                else loss.detach().item()
            )  # the view(-1) seems necessary as f might return a single scalar
            return fval, gradf

    else:
        # This function (that is used if no grads are avail) can also be batched,
        # we just did not batch it so far as the priority is not as high.
        def f_np_wrapper(
            x: npt.NDArray, f: Callable, fixed_features: dict[int, float] | None
        ):
            X = (
                torch.from_numpy(x)
                .to(device=device, dtype=dtype)
                .view(-1, *shapeX[1:])
                .contiguous()
            )
            with torch.no_grad():
                X_fix = fix_features(
                    X=X, fixed_features=fixed_features, replace_current_value=False
                )
                loss = f(X_fix).sum()
            fval = loss.detach().item()
            return fval

    return f_np_wrapper


def get_reasons_against_fast_path(
    method: str,
    with_grad: bool,
    minimize_options: dict[str, Any],
    timeout_sec: float | None,
) -> list[str]:
    # this if-statement is a homage to pytorch's nn.MultiheadAttentionby by @swolchok
    why_not_fast_path = []
    if not method == "L-BFGS-B":
        why_not_fast_path.append(f"method={method}, method needs to be L-BFGS-B")
    if not with_grad:
        why_not_fast_path.append("with_grad=False, it needs to be True")
    if extra_keys := set(minimize_options.keys()) - {
        "maxiter",
        "disp",
        "iprint",
        "max_cor",
        "ftol",
        "pgtol",
        "factr",
        "tol",
        "maxls",
    }:
        why_not_fast_path.append(f"options={extra_keys} are not accepted")
    if timeout_sec is not None:
        why_not_fast_path.append(f"timeout_sec={timeout_sec}, it needs to be None")
    if not check_scipy_version_at_least(minor=13) or check_scipy_version_at_least(
        minor=17
    ):  # pragma: no cover
        # In SciPy 1.15.0, the fortran implementation of L-BFGS-B was
        # translated to C changing its interface slightly.
        # Additionally, we don't know what the future might hold in scipy,
        # thus we use this function to use less optimized code for too new
        # scipy versions.
        why_not_fast_path.append(
            "Scipy version is not in the range from 1.13.0 to 1.15.x."
        )
    return why_not_fast_path


def gen_candidates_torch(
    initial_conditions: Tensor,
    acquisition_function: AcquisitionFunction,
    lower_bounds: float | Tensor | None = None,
    upper_bounds: float | Tensor | None = None,
    optimizer: type[Optimizer] = torch.optim.Adam,
    options: dict[str, float | str] | None = None,
    callback: Callable[[int, Tensor, Tensor], NoReturn] | None = None,
    fixed_features: Mapping[int, float | Tensor] | None = None,
    timeout_sec: float | None = None,
) -> tuple[Tensor, Tensor]:
    r"""Generate a set of candidates using a `torch.optim` optimizer.

    Optimizes an acquisition function starting from a set of initial candidates
    using an optimizer from `torch.optim`.

    Args:
        initial_conditions: Starting points for optimization.
        acquisition_function: Acquisition function to be used.
        lower_bounds: Minimum values for each column of initial_conditions.
        upper_bounds: Maximum values for each column of initial_conditions.
        optimizer (Optimizer): The pytorch optimizer to use to perform
            candidate search.
        options: Options used to control the optimization. Includes
            maxiter: Maximum number of iterations
        callback: A callback function accepting the current iteration, loss,
            and gradients as arguments. This function is executed after computing
            the loss and gradients, but before calling the optimizer.
        fixed_features: This is a dictionary of feature indices to values, where
            all generated candidates will have features fixed to these values.
            If a float is passed it is fixed across [b,q], if a tensor is passed:
            it might either be of shape [b,q] or [b], in which case the same value
            is used across the q dimension.
            Assumes values to be compatible with lower_bounds and upper_bounds!
        timeout_sec: Timeout (in seconds) for optimization. If provided,
            `gen_candidates_torch` will stop after this many seconds and return
            the best solution found so far.

    Returns:
        2-element tuple containing

        - The set of generated candidates.
        - The acquisition value for each t-batch.

    Example:
        >>> qEI = qExpectedImprovement(model, best_f=0.2)
        >>> bounds = torch.tensor([[0., 0.], [1., 2.]])
        >>> Xinit = gen_batch_initial_conditions(
        >>>     qEI, bounds, q=3, num_restarts=25, raw_samples=500
        >>> )
        >>> batch_candidates, batch_acq_values = gen_candidates_torch(
                initial_conditions=Xinit,
                acquisition_function=qEI,
                lower_bounds=bounds[0],
                upper_bounds=bounds[1],
            )
    """
    start_time = time.monotonic()
    options = options or {}
    # We remove max_optimization_problem_aggregation_size as it does not affect
    # the 1st order optimizers implemented in this method.
    # Here, it does not matter whether one combines multiple optimizations into
    # one or not.
    options.pop("max_optimization_problem_aggregation_size", None)
    _clamp = partial(columnwise_clamp, lower=lower_bounds, upper=upper_bounds)
    clamped_candidates = _clamp(initial_conditions)
    if fixed_features:
        clamped_candidates = clamped_candidates[
            ...,
            [i for i in range(clamped_candidates.shape[-1]) if i not in fixed_features],
        ]
    clamped_candidates = clamped_candidates.requires_grad_(True)
    _optimizer = optimizer(params=[clamped_candidates], lr=options.get("lr", 0.025))

    i = 0
    stop = False
    stopping_criterion = ExpMAStoppingCriterion(**options)
    while not stop:
        i += 1
        with torch.no_grad():
            X = _clamp(clamped_candidates).requires_grad_(True)

        loss = -acquisition_function(fix_features(X, fixed_features)).sum()
        grad = torch.autograd.grad(loss, X)[0]
        if callback:
            callback(i, loss, grad)

        def assign_grad():
            _optimizer.zero_grad()
            clamped_candidates.grad = grad
            return loss

        _optimizer.step(assign_grad)
        stop = stopping_criterion(fvals=loss.detach())
        if timeout_sec is not None:
            runtime = time.monotonic() - start_time
            if runtime > timeout_sec:
                stop = True
                logger.info(f"Optimization timed out after {runtime} seconds.")

    clamped_candidates = _clamp(clamped_candidates)
    clamped_candidates = fix_features(clamped_candidates, fixed_features)
    with torch.no_grad():
        batch_acquisition = acquisition_function(clamped_candidates)

    return clamped_candidates, batch_acquisition


def get_best_candidates(batch_candidates: Tensor, batch_values: Tensor) -> Tensor:
    r"""Extract best (q-batch) candidate from batch of candidates

    Args:
        batch_candidates: A `b x q x d` tensor of `b` q-batch candidates, or a
            `b x d` tensor of `b` single-point candidates.
        batch_values: A tensor with `b` elements containing the value of the
            respective candidate (higher is better).

    Returns:
        A tensor of size `q x d` (if q-batch mode) or `d` from batch_candidates
        with the highest associated value.

    Example:
        >>> qEI = qExpectedImprovement(model, best_f=0.2)
        >>> bounds = torch.tensor([[0., 0.], [1., 2.]])
        >>> Xinit = gen_batch_initial_conditions(
        >>>     qEI, bounds, q=3, num_restarts=25, raw_samples=500
        >>> )
        >>> batch_candidates, batch_acq_values = gen_candidates_scipy(
                initial_conditions=Xinit,
                acquisition_function=qEI,
                lower_bounds=bounds[0],
                upper_bounds=bounds[1],
            )
        >>> best_candidates = get_best_candidates(batch_candidates, batch_acq_values)
    """
    best = torch.argmax(batch_values.view(-1), dim=0)
    return batch_candidates[best]


def _process_scipy_result(res: OptimizeResult, options: dict[str, Any]) -> None:
    r"""Process scipy optimization result to produce relevant logs and warnings."""
    if "success" not in res.keys() or "status" not in res.keys():
        with warnings.catch_warnings():
            warnings.simplefilter("always", category=OptimizationWarning)
            warnings.warn(
                "Optimization failed within `scipy.optimize.minimize` with no "
                "status returned to `res.`",
                OptimizationWarning,
                stacklevel=3,
            )
    elif not res.success:
        if (
            "ITERATIONS REACHED LIMIT" in res.message
            or "Iteration limit reached" in res.message
        ):
            logger.info(
                "`scipy.optimize.minimize` exited by reaching the iteration limit of "
                f"`maxiter: {options.get('maxiter')}`."
            )
        elif "EVALUATIONS EXCEEDS LIMIT" in res.message:
            logger.info(
                "`scipy.optimize.minimize` exited by reaching the function evaluation "
                f"limit of `maxfun: {options.get('maxfun')}`."
            )
        elif "Optimization timed out after" in res.message:
            logger.info(res.message)
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("always", category=OptimizationWarning)
                warnings.warn(
                    f"Optimization failed within `scipy.optimize.minimize` with status "
                    f"{res.status} and message {res.message}.",
                    OptimizationWarning,
                    stacklevel=3,
                )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.models.utils.assorted import (
    _make_X_full,
    add_output_dim,
    check_min_max_scaling,
    check_no_nans,
    check_standardization,
    consolidate_duplicates,
    detect_duplicates,
    extract_targets_and_noise_single_output,
    fantasize,
    gpt_posterior_settings,
    mod_batch_shape,
    multioutput_to_batch_mode_transform,
    restore_targets_and_noise_single_output,
    validate_input_scaling,
)


__all__ = [
    "_make_X_full",
    "add_output_dim",
    "check_no_nans",
    "check_min_max_scaling",
    "check_standardization",
    "fantasize",
    "gpt_posterior_settings",
    "multioutput_to_batch_mode_transform",
    "mod_batch_shape",
    "validate_input_scaling",
    "detect_duplicates",
    "consolidate_duplicates",
    "extract_targets_and_noise_single_output",
    "restore_targets_and_noise_single_output",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.optim.utils.acquisition_utils import (
    columnwise_clamp,
    fix_features,
    get_X_baseline,
)
from botorch.optim.utils.common import (
    _handle_numerical_errors,
    _warning_handler_template,
    check_scipy_version_at_least,
)
from botorch.optim.utils.model_utils import (
    get_data_loader,
    get_name_filter,
    get_parameters,
    get_parameters_and_bounds,
    sample_all_priors,
    TorchAttr,
)
from botorch.optim.utils.numpy_utils import as_ndarray, get_bounds_as_ndarray
from botorch.optim.utils.timeout import minimize_with_timeout

__all__ = [
    "_handle_numerical_errors",
    "_warning_handler_template",
    "check_scipy_version_at_least",
    "as_ndarray",
    "columnwise_clamp",
    "fix_features",
    "get_name_filter",
    "get_bounds_as_ndarray",
    "get_data_loader",
    "get_parameters",
    "get_parameters_and_bounds",
    "get_X_baseline",
    "minimize_with_timeout",
    "sample_all_priors",
    "TorchAttr",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch_community.acquisition.bayesian_active_learning import (
    qBayesianQueryByComittee,
    qBayesianVarianceReduction,
    qStatisticalDistanceActiveLearning,
)

# NOTE: This import is needed to register the input constructors.
from botorch_community.acquisition.input_constructors import (  # noqa F401
    acqf_input_constructor,
)
from botorch_community.acquisition.rei import (
    LogRegionalExpectedImprovement,
    qLogRegionalExpectedImprovement,
)
from botorch_community.acquisition.scorebo import qSelfCorrectingBayesianOptimization

__all__ = [
    "LogRegionalExpectedImprovement",
    "qBayesianQueryByComittee",
    "qBayesianVarianceReduction",
    "qLogRegionalExpectedImprovement",
    "qSelfCorrectingBayesianOptimization",
    "qStatisticalDistanceActiveLearning",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Pre-packaged kernels for bayesian optimization, including a Scale/Matern
kernel that is well-suited to low-dimensional high-noise problems, and
a dimension-agnostic RBF kernel without outputscale.

References:

.. [Hvarfner2024vanilla]
    C. Hvarfner, E. O. Hellsten, L. Nardi,
    Vanilla Bayesian Optimization Performs Great in High Dimensions.
    In International Conference on Machine Learning, 2024.
"""

from collections.abc import Sequence
from math import log, sqrt

import torch
from gpytorch.constraints.constraints import GreaterThan
from gpytorch.kernels import MaternKernel, RBFKernel, ScaleKernel
from gpytorch.likelihoods.gaussian_likelihood import GaussianLikelihood
from gpytorch.priors.torch_priors import GammaPrior, LogNormalPrior

MIN_INFERRED_NOISE_LEVEL = 1e-4
SQRT2 = sqrt(2)
SQRT3 = sqrt(3)


def get_matern_kernel_with_gamma_prior(
    ard_num_dims: int, batch_shape: torch.Size | None = None
) -> ScaleKernel:
    r"""Constructs the Scale-Matern kernel that is used by default by
    several models. This uses a Gamma(3.0, 6.0) prior for the lengthscale
    and a Gamma(2.0, 0.15) prior for the output scale.
    """
    return ScaleKernel(
        base_kernel=MaternKernel(
            nu=2.5,
            ard_num_dims=ard_num_dims,
            batch_shape=batch_shape,
            lengthscale_prior=GammaPrior(3.0, 6.0),
        ),
        batch_shape=batch_shape,
        outputscale_prior=GammaPrior(2.0, 0.15),
    )


def get_gaussian_likelihood_with_gamma_prior(
    batch_shape: torch.Size | None = None,
) -> GaussianLikelihood:
    r"""Constructs the GaussianLikelihood that is used by default by
    several models. This uses a Gamma(1.1, 0.05) prior and constrains the
    noise level to be greater than MIN_INFERRED_NOISE_LEVEL (=1e-4).
    """
    batch_shape = torch.Size() if batch_shape is None else batch_shape
    noise_prior = GammaPrior(1.1, 0.05)
    noise_prior_mode = (noise_prior.concentration - 1) / noise_prior.rate
    return GaussianLikelihood(
        noise_prior=noise_prior,
        batch_shape=batch_shape,
        noise_constraint=GreaterThan(
            MIN_INFERRED_NOISE_LEVEL,
            transform=None,
            initial_value=noise_prior_mode,
        ),
    )


def get_gaussian_likelihood_with_lognormal_prior(
    batch_shape: torch.Size | None = None,
) -> GaussianLikelihood:
    """Return Gaussian likelihood with a LogNormal(-4.0, 1.0) prior.
    This prior is based on [Hvarfner2024vanilla]_.

    Args:
        batch_shape: Batch shape for the likelihood.

    Returns:
        GaussianLikelihood with LogNormal(-4.0, 1.0) prior and constrains the
        noise level to be greater than MIN_INFERRED_NOISE_LEVEL (=1e-4).
    """
    batch_shape = torch.Size() if batch_shape is None else batch_shape
    noise_prior = LogNormalPrior(loc=-4.0, scale=1.0)
    return GaussianLikelihood(
        noise_prior=noise_prior,
        batch_shape=batch_shape,
        noise_constraint=GreaterThan(
            MIN_INFERRED_NOISE_LEVEL,
            transform=None,
            initial_value=noise_prior.mode,
        ),
    )


def get_covar_module_with_dim_scaled_prior(
    ard_num_dims: int,
    batch_shape: torch.Size | None = None,
    use_rbf_kernel: bool = True,
    active_dims: Sequence[int] | None = None,
) -> MaternKernel | RBFKernel:
    """Returns an RBF or Matern kernel with priors
    from  [Hvarfner2024vanilla]_.

    Args:
        ard_num_dims: Number of feature dimensions for ARD.
        batch_shape: Batch shape for the covariance module.
        use_rbf_kernel: Whether to use an RBF kernel. If False, uses a Matern kernel.
        active_dims: The set of input dimensions to compute the covariances on.
            By default, the covariance is computed using the full input tensor.
            Set this if you'd like to ignore certain dimensions.

    Returns:
        A Kernel constructed according to the given arguments. The prior is constrained
        to have lengthscales larger than 0.025 for numerical stability.
    """
    base_class = RBFKernel if use_rbf_kernel else MaternKernel
    lengthscale_prior = LogNormalPrior(loc=SQRT2 + log(ard_num_dims) * 0.5, scale=SQRT3)
    base_kernel = base_class(
        ard_num_dims=ard_num_dims,
        batch_shape=batch_shape,
        lengthscale_prior=lengthscale_prior,
        lengthscale_constraint=GreaterThan(
            2.5e-2, transform=None, initial_value=lengthscale_prior.mode
        ),
        # pyre-ignore[6] GPyTorch type is unnecessarily restrictive.
        active_dims=active_dims,
    )
    return base_kernel



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import dataclasses
import itertools
import random
import warnings
from collections.abc import Mapping, Sequence
from typing import Any, Callable

import torch
from botorch.acquisition import AcquisitionFunction
from botorch.exceptions.errors import CandidateGenerationError, UnsupportedError
from botorch.exceptions.warnings import OptimizationWarning
from botorch.generation.gen import gen_candidates_scipy
from botorch.optim.initializers import initialize_q_batch
from botorch.optim.optimize import (
    _optimize_acqf,
    _validate_sequential_inputs,
    OptimizeAcqfInputs,
)
from botorch.optim.parameter_constraints import evaluate_feasibility
from botorch.optim.utils.acquisition_utils import fix_features, get_X_baseline
from botorch.utils.sampling import (
    draw_sobol_samples,
    HitAndRunPolytopeSampler,
    sparse_to_dense_constraints,
)
from botorch.utils.transforms import unnormalize
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor
from torch.quasirandom import SobolEngine

# Default values.
# NOTE: When changing a default, update the corresponding value in the docstrings.
STD_CONT_PERTURBATION = 0.1
RAW_SAMPLES = 1024  # Number of candidates from which to select starting points.
NUM_RESTARTS = 20  # Number of restarts of optimizer with different starting points.
MAX_BATCH_SIZE = 2048  # Maximum batch size.
MAX_ITER_ALTER = 64  # Maximum number of alternating iterations.
MAX_ITER_DISCRETE = 4  # Maximum number of discrete iterations.
MAX_ITER_CONT = 8  # Maximum number of continuous iterations.
# Maximum number of discrete values for a discrete dimension.
# If there are more values for a dimension, we will use continuous
# relaxation to optimize it.
MAX_DISCRETE_VALUES = 20
# Maximum number of iterations for optimizing the continuous relaxation
# during initialization
MAX_ITER_INIT = 100
CONVERGENCE_TOL = 1e-8  # Optimizer convergence tolerance.
DUPLICATE_TOL = 1e-6  # Tolerance for deduplicating initial candidates.
STOP_AFTER_SHARE_CONVERGED = 1.0  # We optimize multiple configurations at once
# in `optimize_acqf_mixed_alternating`. This option controls, whether to stop
# optimizing after the given share has converged.
# Convergence is defined as the improvements of one discrete, followed by a scalar
# optimization yield less than `CONVERGENCE_TOL` improvements.

SUPPORTED_OPTIONS = {
    "initialization_strategy",
    "tol",
    "maxiter_alternating",
    "maxiter_discrete",
    "maxiter_continuous",
    "maxiter_init",
    "max_discrete_values",
    "num_spray_points",
    "std_cont_perturbation",
    "batch_limit",
    "init_batch_limit",
}
SUPPORTED_INITIALIZATION = {"continuous_relaxation", "equally_spaced", "random"}


def _setup_continuous_relaxation(
    discrete_dims: dict[int, list[float]],
    max_discrete_values: int,
    post_processing_func: Callable[[Tensor], Tensor] | None,
) -> tuple[list[int], Callable[[Tensor], Tensor] | None]:
    r"""Update `discrete_dims` and `post_processing_func` to use
    continuous relaxation for discrete dimensions that have more than
    `max_discrete_values` values. These dimensions are removed from
    `discrete_dims` and `post_processing_func` is updated to round
    them to the nearest integer.
    """

    dims_to_relax, dims_to_keep = {}, {}
    for index, values in discrete_dims.items():
        if len(values) > max_discrete_values:
            dims_to_relax[index] = values
        else:
            dims_to_keep[index] = values

    if len(dims_to_relax) == 0:
        return discrete_dims, post_processing_func

    def new_post_processing_func(X: Tensor) -> Tensor:
        r"""Round the relaxed dimensions to the nearest integer and apply the original
        `post_processing_func`."""

        X = round_discrete_dims(X=X, discrete_dims=dims_to_relax)
        if post_processing_func is not None:
            X = post_processing_func(X)
        return X

    return dims_to_keep, new_post_processing_func


def _filter_infeasible(
    X: Tensor,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None,
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None,
) -> Tensor:
    r"""Filters infeasible points from a set of points.

    NOTE: This function only supports intra-point constraints. This is validated
        in `optimize_acqf_mixed_alternating`, so we do not repeat the
        validation in here.

    Args:
        X: A tensor of points of shape `n x d`.
        inequality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) >= rhs`. `indices` and
            `coefficients` should be torch tensors. See the docstring of
            `make_scipy_linear_constraints` for an example.
        equality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an equality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) == rhs`. `indices` and
            `coefficients` should be torch tensors. Example:
            `[(torch.tensor([1, 3]), torch.tensor([1.0, 0.5]), -0.1)]`

    Returns:
        The tensor `X` with infeasible points removed.
    """
    # X is reshaped to [n, 1, d] in order to be able to apply
    # `evaluate_feasibility` which operates on the batch level
    is_feasible = evaluate_feasibility(
        X=X.unsqueeze(-2),
        inequality_constraints=inequality_constraints,
        equality_constraints=equality_constraints,
        nonlinear_inequality_constraints=None,
    )
    return X[is_feasible]


def get_nearest_neighbors(
    current_x: Tensor,
    bounds: Tensor,
    discrete_dims: dict[int, list[float]],
) -> Tensor:
    r"""Generate all 1-Manhattan distance neighbors of a given input. The neighbors
    are generated for the discrete dimensions only.

    NOTE: This assumes that `current_x` is detached and uses in-place operations,
    which are known to be incompatible with autograd.

    Args:
        current_x: The design to find the neighbors of. A tensor of shape `d`.
        bounds: A `2 x d` tensor of lower and upper bounds for each column of `X`.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.

    Returns:
        A tensor of shape `num_neighbors x d`, denoting all unique 1-Manhattan
        distance neighbors.
    """
    # we first transform the discrete values in the current_x to its index in the
    # values of the discrete_dims, so that we can easily increase/decrease
    # the discrete dimensions by one.
    # Map discrete values in current_x to their indices in discrete_dims
    discrete_indices = []

    t_discrete_dims = torch.tensor(
        list(discrete_dims.keys()), dtype=torch.long, device=current_x.device
    )

    for dim, values in discrete_dims.items():
        val = current_x[dim].item()
        # Find the index of the closest value in values
        idx = min(range(len(values)), key=lambda i: abs(values[i] - val))
        discrete_indices.append(idx)
    current_x_int = current_x.clone()
    current_x_int[t_discrete_dims] = torch.tensor(
        discrete_indices, dtype=current_x.dtype, device=current_x.device
    )
    lower_clamp = bounds[0].clone()
    lower_clamp[t_discrete_dims] = torch.tensor(
        [0 for _ in discrete_dims.values()],
        dtype=lower_clamp.dtype,
        device=lower_clamp.device,
    )
    upper_clamp = bounds[1].clone()
    upper_clamp[t_discrete_dims] = torch.tensor(
        [len(values) - 1 for values in discrete_dims.values()],
        dtype=upper_clamp.dtype,
        device=upper_clamp.device,
    )
    num_discrete = len(discrete_dims)
    diag_ones = torch.eye(num_discrete, dtype=current_x.dtype, device=current_x.device)
    # Neighbors obtained by increasing a discrete dimension by one.
    plus_neighbors = current_x_int.repeat(num_discrete, 1)
    plus_neighbors[:, t_discrete_dims] += diag_ones
    plus_neighbors.clamp_(max=upper_clamp)
    # Neighbors obtained by decreasing a discrete dimension by one.
    minus_neighbors = current_x_int.repeat(num_discrete, 1)
    minus_neighbors[:, t_discrete_dims] -= diag_ones
    minus_neighbors.clamp_(min=lower_clamp)
    unique_neighbors = torch.cat([minus_neighbors, plus_neighbors], dim=0).unique(dim=0)
    # Also remove current_x if it is in unique_neighbors.
    unique_neighbors = unique_neighbors[
        ~(unique_neighbors == current_x_int).all(dim=-1)
    ]

    # Transform unique_neighbors back to the original non-integer space
    for dim, values in discrete_dims.items():
        t_values = torch.tensor(
            values, device=unique_neighbors.device, dtype=unique_neighbors.dtype
        )
        idx = unique_neighbors[:, dim].long()
        unique_neighbors[:, dim] = t_values[idx]

    return unique_neighbors


def get_categorical_neighbors(
    current_x: Tensor,
    cat_dims: dict[int, list[float]],
    max_num_cat_values: int = MAX_DISCRETE_VALUES,
) -> Tensor:
    r"""Generate all 1-Hamming distance neighbors of a given input. The neighbors
    are generated for the categorical dimensions only.

    We assume that all categorical values are equidistant. If the number of values
    is greater than `max_num_cat_values`, we sample uniformly from the
    possible values for that dimension.

    NOTE: This assumes that `current_x` is detached and uses in-place operations,
    which are known to be incompatible with autograd.

    Args:
        current_x: The design to find the neighbors of. A tensor of shape `d`.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        max_num_cat_values: Maximum number of values for a categorical parameter,
            beyond which values are uniformly sampled.

    Returns:
        A tensor of shape `num_neighbors x d`, denoting up to `max_num_cat_values`
        unique 1-Hamming distance neighbors for each categorical dimension.
    """

    # Neighbors are generated by considering all possible values for each
    # categorical dimension, one at a time.
    def _get_cat_values(dim: int) -> Sequence[int]:
        r"""Get a sequence of up to `max_num_cat_values` values that a categorical
        feature may take."""
        current_value = current_x[dim]
        if len(cat_dims[dim]) <= max_num_cat_values:
            return cat_dims[dim]
        else:
            return random.sample(
                [v for v in cat_dims[dim] if v != current_value], k=max_num_cat_values
            )

    new_cat_values_dict = {dim: _get_cat_values(dim) for dim in cat_dims.keys()}
    new_cat_values_lst = list(
        itertools.chain.from_iterable(new_cat_values_dict.values())
    )
    new_cat_values = torch.tensor(
        new_cat_values_lst, device=current_x.device, dtype=current_x.dtype
    )

    new_cat_idcs = torch.cat(
        tuple(
            torch.full(
                (min(len(values), max_num_cat_values),), dim, device=current_x.device
            )
            for dim, values in new_cat_values_dict.items()
        )
    )

    neighbors = current_x.repeat(len(new_cat_values), 1)
    # Assign the new values to their corresponding columns.
    neighbors.scatter_(1, new_cat_idcs.view(-1, 1), new_cat_values.view(-1, 1))

    unique_neighbors = neighbors.unique(dim=0)
    # Also remove current_x if it is in unique_neighbors.
    unique_neighbors = unique_neighbors[~(unique_neighbors == current_x).all(dim=-1)]
    return unique_neighbors


def get_spray_points(
    X_baseline: Tensor,
    cont_dims: Tensor,
    discrete_dims: dict[int, list[float]],
    cat_dims: dict[int, list[float]],
    bounds: Tensor,
    num_spray_points: int,
    std_cont_perturbation: float = STD_CONT_PERTURBATION,
) -> Tensor:
    r"""Generate spray points by perturbing the Pareto optimal points.

    Given the points on the Pareto frontier, we create perturbations (spray points)
    by adding Gaussian perturbation to the continuous parameters and 1-Manhattan
    distance neighbors of the discrete (binary and integer) parameters.

    Args:
        X_baseline: Tensor of best acquired points across BO run.
        cont_dims: Indices of continuous parameters/input dimensions.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        bounds: A `2 x d` tensor of lower and upper bounds for each column of `X`.
        num_spray_points: Number of spray points to return.
        std_cont_perturbation: standard deviation of Normal perturbations of
            continuous dimensions. Default is STD_CONT_PERTURBATION = 0.2.

    Returns:
        A (num_spray_points x d)-dim tensor of perturbed points.
    """
    dim = bounds.shape[-1]
    device, dtype = X_baseline.device, X_baseline.dtype
    perturb_nbors = torch.zeros(0, dim, device=device, dtype=dtype)
    t_discrete_dims = torch.tensor(
        list(discrete_dims.keys()), dtype=torch.long, device=device
    )
    t_cat_dims = torch.tensor(list(cat_dims.keys()), dtype=torch.long, device=device)
    for x in X_baseline:
        if len(discrete_dims) > 0:
            discrete_perturbs = get_nearest_neighbors(
                current_x=x, bounds=bounds, discrete_dims=discrete_dims
            )
            discrete_perturbs = discrete_perturbs[
                torch.randint(
                    len(discrete_perturbs), (num_spray_points,), device=device
                )
            ]
        if len(cat_dims) > 0:
            cat_perturbs = get_categorical_neighbors(current_x=x, cat_dims=cat_dims)
            cat_perturbs = cat_perturbs[
                torch.randint(len(cat_perturbs), (num_spray_points,), device=device)
            ]

        cont_perturbs = x[cont_dims] + std_cont_perturbation * torch.randn(
            num_spray_points, len(cont_dims), device=device, dtype=dtype
        )
        cont_perturbs = cont_perturbs.clamp_(
            min=bounds[0, cont_dims], max=bounds[1, cont_dims]
        )
        nbds = torch.zeros(num_spray_points, dim, device=device, dtype=dtype)
        if len(discrete_dims) > 0:
            nbds[..., t_discrete_dims] = discrete_perturbs[..., t_discrete_dims]
        if len(cat_dims) > 0:
            nbds[..., t_cat_dims] = cat_perturbs[..., t_cat_dims]

        nbds[..., cont_dims] = cont_perturbs
        perturb_nbors = torch.cat([perturb_nbors, nbds], dim=0)
    return perturb_nbors


def sample_feasible_points(
    opt_inputs: OptimizeAcqfInputs,
    discrete_dims: dict[int, list[float]],
    cat_dims: dict[int, list[float]],
    num_points: int,
) -> Tensor:
    r"""Sample feasible points from the optimization domain.

    Feasibility is determined according to the discrete dimensions taking
    integer values and the inequality constraints being satisfied.

    If there are no inequality constraints, Sobol is used to generate the base points.
    Otherwise, we use the polytope sampler to generate the base points. The base points
    are then rounded to the nearest integer values for the discrete dimensions, and
    the infeasible points are filtered out (in case rounding leads to infeasibility).

    This method will do 10 attempts to generate `num_points` feasible points, and
    return the points generated so far. If no points are generated, it will error out.

    Args:
        opt_inputs: Common set of arguments for acquisition optimization.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        num_points: The number of points to sample.

    Returns:
        A tensor of shape `num_points x d` containing the sampled points.
    """
    bounds = opt_inputs.bounds
    all_points = torch.empty(
        0, bounds.shape[-1], device=bounds.device, dtype=bounds.dtype
    )
    if (
        opt_inputs.inequality_constraints is None
        and opt_inputs.equality_constraints is None
    ):
        # Generate base points using Sobol.
        sobol_engine = SobolEngine(dimension=bounds.shape[-1], scramble=True)

        def generator(n: int) -> Tensor:
            samples = sobol_engine.draw(n=n, dtype=bounds.dtype).to(bounds.device)
            return unnormalize(X=samples, bounds=bounds)

    else:
        # Generate base points using polytope sampler.
        # Since we may generate many times, we initialize the sampler with burn-in
        # to reduce the start-up cost for subsequent calls.
        if opt_inputs.inequality_constraints is not None:
            A, b = sparse_to_dense_constraints(
                d=bounds.shape[-1], constraints=opt_inputs.inequality_constraints
            )
            ineqs = (-A, -b)
        else:
            ineqs = None
        if opt_inputs.equality_constraints is not None:
            A_eq, b_eq = sparse_to_dense_constraints(
                d=bounds.shape[-1], constraints=opt_inputs.equality_constraints
            )
            eqs = (A_eq, b_eq)
        else:
            eqs = None
        polytope_sampler = HitAndRunPolytopeSampler(
            bounds=bounds, inequality_constraints=ineqs, equality_constraints=eqs
        )

        def generator(n: int) -> Tensor:
            return polytope_sampler.draw(n=n)

    for _ in range(10):
        num_remaining = num_points - len(all_points)
        if num_remaining <= 0:
            break
        # Generate twice as many, since we're likely to filter out some points.
        base_points = generator(n=num_remaining * 2)
        # Round the discrete dimensions to the nearest integer.
        base_points = round_discrete_dims(X=base_points, discrete_dims=discrete_dims)
        base_points = round_discrete_dims(X=base_points, discrete_dims=cat_dims)
        # Fix the fixed features.
        base_points = fix_features(
            X=base_points,
            fixed_features=opt_inputs.fixed_features,
            replace_current_value=True,
        )
        # Filter out infeasible points.
        feasible_points = _filter_infeasible(
            X=base_points,
            inequality_constraints=opt_inputs.inequality_constraints,
            equality_constraints=opt_inputs.equality_constraints,
        )
        all_points = torch.cat([all_points, feasible_points], dim=0)

    if len(all_points) == 0:
        raise CandidateGenerationError(
            "Could not generate any feasible starting points for mixed optimizer."
        )
    return all_points[:num_points]


def round_discrete_dims(X: Tensor, discrete_dims: dict[int, list[float]]) -> Tensor:
    """Round the discrete dimensions of a tensor to the nearest allowed values.

    Args:
        X: A tensor of shape `n x d`, where `d` is the number of dimensions.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.

    Returns:
        A tensor of the same shape as `X`, with discrete dimensions rounded to
        the nearest allowed values.
    """
    X = X.clone()
    for dim, values in discrete_dims.items():
        allowed = torch.tensor(values, device=X.device, dtype=X.dtype)
        diffs = torch.abs(X[..., dim].unsqueeze(-1) - allowed)
        idx = torch.argmin(diffs, dim=-1)
        X[..., dim] = allowed[idx]
    return X


def generate_starting_points(
    opt_inputs: OptimizeAcqfInputs,
    discrete_dims: dict[int, list[float]],
    cat_dims: dict[int, list[float]],
    cont_dims: Tensor,
) -> tuple[Tensor, Tensor]:
    """Generate initial starting points for the alternating optimization.

    This method attempts to generate the initial points using the specified
    options and completes any missing points using `sample_feasible_points`.

    Args:
        opt_inputs: Common set of arguments for acquisition optimization.
            This function utilizes `acq_function`, `bounds`, `num_restarts`,
            `raw_samples`, `options`, `fixed_features` and constraints
            from `opt_inputs`.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        cont_dims: A tensor of indices corresponding to continuous parameters.

    Returns:
        A tuple of two tensors: a (num_restarts x d)-dim tensor of starting points
        and a (num_restarts)-dim tensor of their respective acquisition values.
        In rare cases, this method may return fewer than `num_restarts` points.
    """
    bounds = opt_inputs.bounds
    binary_dims = []
    for dim, values in discrete_dims.items():
        if len(values) == 2:
            binary_dims.append(dim)
    num_binary = len(binary_dims)
    num_discrete = len(discrete_dims) - num_binary
    num_restarts = opt_inputs.num_restarts
    raw_samples = none_throws(opt_inputs.raw_samples)

    options = opt_inputs.options or {}
    initialization_strategy = options.get(
        "initialization_strategy",
        (
            "equally_spaced"
            if num_discrete == 0 and num_binary >= 2
            else "continuous_relaxation"
        ),
    )
    if initialization_strategy not in SUPPORTED_INITIALIZATION:
        raise UnsupportedError(  # pragma: no cover
            f"Unsupported initialization strategy: {initialization_strategy}."
            f"Supported strategies are: {SUPPORTED_INITIALIZATION}."
        )

    # Initialize `x_init_candts` here so that it's always defined as a tensor.
    x_init_candts = torch.empty(
        0, bounds.shape[-1], device=bounds.device, dtype=bounds.dtype
    )
    if initialization_strategy == "continuous_relaxation":
        try:
            # Optimize the acquisition function with continuous relaxation.
            updated_opt_inputs = dataclasses.replace(
                opt_inputs,
                q=1,
                return_best_only=False,
                options={
                    "maxiter": options.get("maxiter_init", MAX_ITER_INIT),
                    "batch_limit": options.get("batch_limit", MAX_BATCH_SIZE),
                    "init_batch_limit": options.get("init_batch_limit", MAX_BATCH_SIZE),
                },
                post_processing_func=None,
            )
            x_init_candts, _ = _optimize_acqf(opt_inputs=updated_opt_inputs)
            x_init_candts = x_init_candts.squeeze(-2).detach()
        except Exception as e:
            warnings.warn(
                "Failed to initialize using continuous relaxation. Using "
                "`sample_feasible_points` for initialization. Original error "
                f"message: {e}",
                OptimizationWarning,
                stacklevel=2,
            )

    if len(x_init_candts) == 0:
        # Generate Sobol points as a fallback for `continuous_relaxation` and for
        # further refinement in `equally_spaced` strategy.
        x_init_candts = draw_sobol_samples(bounds=bounds, n=raw_samples, q=1)
        x_init_candts = x_init_candts.squeeze(-2)

    if initialization_strategy == "equally_spaced":
        if num_discrete > 0:
            raise ValueError(  # pragma: no cover
                "Equally spaced initialization is not supported with non-binary "
                "discrete variables."
            )
        # Picking initial points by equally spaced number of features/binary inputs.
        k = torch.randint(
            low=0,
            high=num_binary,
            size=(raw_samples,),
            dtype=torch.int64,
            device=bounds.device,
        )
        x_init_candts[:, binary_dims] = bounds[0, binary_dims]
        binary_dims_t = torch.as_tensor(binary_dims, device=bounds.device)
        for i, xi in enumerate(x_init_candts):
            rand_binary_dims = binary_dims_t[
                torch.randperm(num_binary, device=xi.device)[: k[i]]
            ]
            x_init_candts[i, rand_binary_dims] = bounds[1, rand_binary_dims]

    num_spray_points = assert_is_instance(
        options.get("num_spray_points", 20 if num_discrete == 0 else 0), int
    )
    if (
        num_spray_points > 0
        and (X_baseline := get_X_baseline(acq_function=opt_inputs.acq_function))
        is not None
    ):
        perturb_nbors = get_spray_points(
            X_baseline=X_baseline,
            cont_dims=cont_dims,
            discrete_dims=discrete_dims,
            cat_dims=cat_dims,
            bounds=bounds,
            num_spray_points=num_spray_points,
            std_cont_perturbation=assert_is_instance(
                options.get("std_cont_perturbation", STD_CONT_PERTURBATION), float
            ),
        )
        x_init_candts = torch.cat([x_init_candts, perturb_nbors], dim=0)

    # For each discrete dimension, map to the nearest allowed value
    x_init_candts = round_discrete_dims(X=x_init_candts, discrete_dims=discrete_dims)
    x_init_candts = round_discrete_dims(X=x_init_candts, discrete_dims=cat_dims)
    x_init_candts = fix_features(
        X=x_init_candts,
        fixed_features=opt_inputs.fixed_features,
        replace_current_value=True,
    )
    x_init_candts = _filter_infeasible(
        X=x_init_candts,
        inequality_constraints=opt_inputs.inequality_constraints,
        equality_constraints=opt_inputs.equality_constraints,
    )

    # If there are fewer than `num_restarts` feasible points, attempt to generate more.
    if len(x_init_candts) < num_restarts:
        new_x_init = sample_feasible_points(
            opt_inputs=opt_inputs,
            discrete_dims=discrete_dims,
            cat_dims=cat_dims,
            num_points=num_restarts - len(x_init_candts),
        )
        x_init_candts = torch.cat([x_init_candts, new_x_init], dim=0)

    with torch.no_grad():
        acq_vals = torch.cat(
            [
                opt_inputs.acq_function(X_.unsqueeze(-2))
                for X_ in x_init_candts.split(
                    options.get("init_batch_limit", MAX_BATCH_SIZE)
                )
            ]
        )
    if len(x_init_candts) > num_restarts:
        # If there are more than `num_restarts` feasible points, select a diverse
        # set of initializers using Boltzmann sampling.
        x_init_candts, acq_vals = initialize_q_batch(
            X=x_init_candts, acq_vals=acq_vals, n=num_restarts
        )

    return x_init_candts, acq_vals


def discrete_step(
    opt_inputs: OptimizeAcqfInputs,
    discrete_dims: dict[int, list[float]],
    cat_dims: dict[int, list[float]],
    current_x: Tensor,
) -> tuple[Tensor, Tensor]:
    """Discrete nearest neighbour search.

    Args:
        opt_inputs: Common set of arguments for acquisition optimization.
            This function utilizes `acq_function`, `bounds`, `options`
            and constraints from `opt_inputs`.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        current_x: Batch of starting points. A tensor of shape `b x d`.

    Returns:
        A tuple of two tensors: a (b, d)-dim tensor of optimized point
            and a scalar tensor of correspondins acquisition value.
    """
    with torch.no_grad():
        current_acqvals = opt_inputs.acq_function(current_x.unsqueeze(1))
    options = opt_inputs.options or {}
    maxiter_discrete = options.get("maxiter_discrete", MAX_ITER_DISCRETE)
    done = torch.zeros(len(current_x), dtype=torch.bool)
    for _ in range(assert_is_instance(maxiter_discrete, int)):
        # we don't batch this, as the number of x_neighbors can be different
        # for each entry (as duplicates are removed), and the most expensive
        # op is the acq_function, which is batched
        # TODO finding the set of neighbors currently is done sequentially
        # for one item in the batch after the other
        x_neighbors_list = [None for _ in done]
        for i in range(len(done)):
            # don't do anything if we are already done
            if done[i]:
                continue

            neighbors = []

            # if we have discrete_dims look for neighbors by stepping +1/-1
            if len(discrete_dims) > 0:
                x_neighbors_discrete = get_nearest_neighbors(
                    current_x=current_x[i].detach(),
                    bounds=opt_inputs.bounds,
                    discrete_dims=discrete_dims,
                )
                x_neighbors_discrete = _filter_infeasible(
                    X=x_neighbors_discrete,
                    inequality_constraints=opt_inputs.inequality_constraints,
                    equality_constraints=opt_inputs.equality_constraints,
                )
                neighbors.append(x_neighbors_discrete)

            # if we have cat_dims look for neighbors by changing the cat's
            if len(cat_dims) > 0:
                x_neighbors_cat = get_categorical_neighbors(
                    current_x=current_x[i].detach(),
                    cat_dims=cat_dims,
                )
                x_neighbors_cat = _filter_infeasible(
                    X=x_neighbors_cat,
                    inequality_constraints=opt_inputs.inequality_constraints,
                    equality_constraints=opt_inputs.equality_constraints,
                )
                neighbors.append(x_neighbors_cat)

            x_neighbors = torch.cat(neighbors, dim=0)
            if x_neighbors.numel() == 0:
                # If the i'th point has no neighbors, we mark it as done
                done[i] = True  # pragma: no cover
            x_neighbors_list[i] = x_neighbors

        # Exit if all batches converged or have no valid neighbors left.
        if done.all():
            break

        all_x_neighbors = torch.cat(
            [
                x_neighbors
                for x_neighbors in x_neighbors_list
                if x_neighbors is not None
            ],
            dim=0,
        )  # shape: (sum(#neihbors of the items in the batch), d)
        with torch.no_grad():
            # This is the most expensive call in this function.
            # The reason that `discrete_step` uses a batched x
            # rather than looping over each batch within x is so that
            # we can batch this call. This leads to an overall speedup
            # even though the above and below for loops cannot
            # be sped up by batching.
            acq_vals = torch.cat(
                [
                    opt_inputs.acq_function(X_.unsqueeze(-2))
                    for X_ in all_x_neighbors.split(
                        options.get("init_batch_limit", MAX_BATCH_SIZE)
                    )
                ]
            )
        offset = 0
        for i in range(len(done)):
            if done[i]:
                continue

            # We index into all_x_neighbors in the following convoluted way,
            # as it is a flattened version of x_neighbors_list with the
            # None entries removed. That is why we do not increase offset if done[i].
            width = len(x_neighbors_list[i])
            x_neighbors = all_x_neighbors[offset : offset + width]
            max_acq, argmax = acq_vals[offset : offset + width].max(dim=0)
            improvement = max_acq - current_acqvals[i]
            if improvement > 0:
                current_acqvals[i], current_x[i] = (
                    max_acq,
                    x_neighbors[argmax],
                )
            if improvement <= options.get("tol", CONVERGENCE_TOL):
                done[i] = True

            offset += width

    return current_x, current_acqvals


def continuous_step(
    opt_inputs: OptimizeAcqfInputs,
    discrete_dims: Tensor,
    cat_dims: Tensor,
    current_x: Tensor,
) -> tuple[Tensor, Tensor]:
    """Continuous search using L-BFGS-B through optimize_acqf.

    Args:
        opt_inputs: Common set of arguments for acquisition optimization.
            This function utilizes `acq_function`, `bounds`, `options`,
            `fixed_features` and constraints from `opt_inputs`.
            `opt_inputs.return_best_only` should be `False`.
        discrete_dims: A dictionary mapping indices of discrete dimensions
            to a list of allowed values for that dimension.
        cat_dims: A tensor of indices corresponding to categorical parameters.
        current_x: Starting point. A tensor of shape `b x d`.

    Returns:
        A tuple of two tensors: a (b x d)-dim tensor of optimized points
            and a (b)-dim tensor of acquisition values.
    """

    if opt_inputs.return_best_only:
        raise UnsupportedError(
            "`continuous_step` does not support `return_best_only=True`."
        )

    d = current_x.shape[-1]
    options = opt_inputs.options or {}
    non_cont_dims = torch.cat((discrete_dims, cat_dims), dim=0)

    if len(non_cont_dims) == d:  # nothing continuous to optimize
        with torch.no_grad():
            return current_x, opt_inputs.acq_function(current_x.unsqueeze(1))

    updated_opt_inputs = dataclasses.replace(
        opt_inputs,
        q=1,
        raw_samples=None,
        # unsqueeze to add the q dimension
        batch_initial_conditions=current_x.unsqueeze(1),
        fixed_features={
            **{d: current_x[:, d] for d in non_cont_dims.tolist()},
            **(opt_inputs.fixed_features or {}),
        },
        options={
            "maxiter": options.get("maxiter_continuous", MAX_ITER_CONT),
            "tol": options.get("tol", CONVERGENCE_TOL),
            "batch_limit": options.get("batch_limit", MAX_BATCH_SIZE),
            "max_optimization_problem_aggregation_size": 1,
        },
    )
    best_X, best_acq_values = _optimize_acqf(opt_inputs=updated_opt_inputs)
    return best_X.view_as(current_x), best_acq_values


def optimize_acqf_mixed_alternating(
    acq_function: AcquisitionFunction,
    bounds: Tensor,
    discrete_dims: Mapping[int, Sequence[float]] | None = None,
    cat_dims: Mapping[int, Sequence[float]] | None = None,
    options: dict[str, Any] | None = None,
    q: int = 1,
    raw_samples: int = RAW_SAMPLES,
    num_restarts: int = NUM_RESTARTS,
    post_processing_func: Callable[[Tensor], Tensor] | None = None,
    sequential: bool = True,
    fixed_features: dict[int, float] | None = None,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
) -> tuple[Tensor, Tensor]:
    r"""
    Optimizes acquisition function over mixed integer, categorical, and continuous
    input spaces. Multiple random restarting starting points are picked by evaluating
    a large set of initial candidates. From each starting point, alternating
    discrete/categorical local search and continuous optimization via (L-BFGS)
    is performed for a fixed number of iterations.

    The discrete dimensions that have more than
    `options.get("max_discrete_values", MAX_DISCRETE_VALUES)` values will
    be optimized using continuous relaxation.
    The categorical dimensions that have more than `MAX_DISCRETE_VALUES` values
    be optimized by selecting random subsamples of the possible values.

    Args:
        acq_function: BoTorch Acquisition function.
        bounds: A `2 x d` tensor of lower and upper bounds for each column of `X`.
        discrete_dims: A dictionary mapping indices of discrete and binary
            dimensions to a list of allowed values for that dimension.
        cat_dims: A dictionary mapping indices of categorical dimensions
            to a list of allowed values for that dimension.
        options: Dictionary specifying optimization options. Supports the following:
        - "initialization_strategy": Strategy used to generate the initial candidates.
            "random", "continuous_relaxation" or "equally_spaced" (linspace style).
        - "tol": The algorithm terminates if the absolute improvement in acquisition
            value of one iteration is smaller than this number.
        - "maxiter_alternating": Number of alternating steps. Defaults to 64.
        - "maxiter_discrete": Maximum number of iterations in each discrete step.
            Defaults to 4.
        - "maxiter_continuous": Maximum number of iterations in each continuous step.
            Defaults to 8.
        - "max_discrete_values": Maximum number of values for a discrete dimension
            to be optimized using discrete step / local search. The discrete dimensions
            with more values will be optimized using continuous relaxation.
        - "num_spray_points": Number of spray points (around `X_baseline`) to add to
            the points generated by the initialization strategy. Defaults to 20 if
            all discrete variables are binary and to 0 otherwise.
        - "std_cont_perturbation": Standard deviation of the normal perturbations of
            the continuous variables used to generate the spray points.
            Defaults to 0.1.
        - "batch_limit": The maximum batch size for jointly evaluating candidates
            during optimization.
        - "init_batch_limit": The maximum batch size for jointly evaluating candidates
            during initialization. During initialization, candidates are evaluated
            in a `no_grad` context, which reduces memory usage. As a result,
            `init_batch_limit` can be set to a larger value than `batch_limit`.
            Defaults to `batch_limit`, if given.
        q: Number of candidates.
        raw_samples: Number of initial candidates used to select starting points from.
            Defaults to 1024.
        num_restarts: Number of random restarts. Defaults to 20.
        post_processing_func: A function that post-processes an optimization result
            appropriately (i.e., according to `round-trip` transformations).
        sequential: Whether to use joint or sequential optimization across q-batch.
            This currently only supports sequential optimization.
        fixed_features: A map `{feature_index: value}` for features that
            should be fixed to a particular value during generation.
        inequality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) >= rhs`. `indices` and
            `coefficients` should be torch tensors. See the docstring of
            `make_scipy_linear_constraints` for an example.
        equality_constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an equality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) == rhs`. `indices` and
            `coefficients` should be torch tensors. Example:
            `[(torch.tensor([1, 3]), torch.tensor([1.0, 0.5]), -0.1)]` Equality
            constraints can only be used with continuous degrees of freedom.

    Returns:
        A tuple of two tensors: a (q x d)-dim tensor of optimized points
            and a (q)-dim tensor of their respective acquisition values.
    """

    if sequential is False:  # pragma: no cover
        raise NotImplementedError(
            "`optimize_acqf_mixed_alternating` only supports "
            "sequential optimization."
        )

    cat_dims = cat_dims or {}
    discrete_dims = discrete_dims or {}

    # sort the values in discrete dims in ascending order
    discrete_dims = {dim: sorted(values) for dim, values in discrete_dims.items()}

    # sort the categorical dims in ascending order
    cat_dims = {dim: sorted(values) for dim, values in cat_dims.items()}

    for dim, values in discrete_dims.items():
        lower_bnd, upper_bnd = bounds[:, dim].tolist()
        lower, upper = values[0], values[-1]
        if lower != lower_bnd:
            raise ValueError(
                f"Discrete dimension {dim} must start at the lower bound "
                f"{lower_bnd} but starts at {lower}."
            )
        if upper != upper_bnd:
            raise ValueError(
                f"Discrete dimension {dim} must end at the upper bound "
                f"{upper_bnd} but end at {upper}."
            )

    fixed_features = fixed_features or {}
    options = options or {}
    if options.get("max_optimization_problem_aggregation_size", 1) != 1:
        raise UnsupportedError(
            "optimize_acqf_mixed_alternating does not support "
            "max_optimization_problem_aggregation_size != 1. "
            "You might leave this option empty, though."
        )
    options.setdefault("batch_limit", MAX_BATCH_SIZE)
    options.setdefault("init_batch_limit", options["batch_limit"])
    if not (keys := set(options.keys())).issubset(SUPPORTED_OPTIONS):
        unsupported_keys = keys.difference(SUPPORTED_OPTIONS)
        raise UnsupportedError(
            f"Received an unsupported option {unsupported_keys}. {SUPPORTED_OPTIONS=}."
        )

    if equality_constraints is not None:
        for indices, _, __ in equality_constraints:
            # Raise an error if any index in indices is in discrete_dims or cat_dims
            if any(idx in discrete_dims or idx in cat_dims for idx in indices.tolist()):
                raise ValueError(
                    "Equality constraints can only be used with continuous degrees "
                    "of freedom."
                )

    # Update discrete dims and post processing functions to account for any
    # dimensions that should be using continuous relaxation.
    discrete_dims, post_processing_func = _setup_continuous_relaxation(
        discrete_dims=discrete_dims,
        max_discrete_values=assert_is_instance(
            options.get("max_discrete_values", MAX_DISCRETE_VALUES), int
        ),
        post_processing_func=post_processing_func,
    )

    opt_inputs = OptimizeAcqfInputs(
        acq_function=acq_function,
        bounds=bounds,
        q=q,
        num_restarts=num_restarts,
        raw_samples=raw_samples,
        options=options,
        inequality_constraints=inequality_constraints,
        equality_constraints=equality_constraints,
        nonlinear_inequality_constraints=None,
        fixed_features=fixed_features,
        post_processing_func=post_processing_func,
        batch_initial_conditions=None,
        return_best_only=False,  # We don't want to perform the cont. optimization
        # step and only return best, but this function itself only returns best
        gen_candidates=gen_candidates_scipy,
        sequential=sequential,  # only relevant if all dims are cont.
    )
    if sequential:
        # Sequential optimization requires return_best_only to be True
        # But we turn it off here, as we "manually" perform the seq.
        # conditioning in the loop below
        _validate_sequential_inputs(
            opt_inputs=dataclasses.replace(opt_inputs, return_best_only=True)
        )

    base_X_pending = acq_function.X_pending if q > 1 else None
    dim = bounds.shape[-1]
    tkwargs: dict[str, Any] = {"device": bounds.device, "dtype": bounds.dtype}
    # Remove fixed features from dims, so they don't get optimized.
    discrete_dims = {
        dim: values
        for dim, values in discrete_dims.items()
        if dim not in fixed_features
    }
    cat_dims = {
        dim: values for dim, values in cat_dims.items() if dim not in fixed_features
    }
    non_cont_dims = [*discrete_dims.keys(), *cat_dims.keys()]
    if len(non_cont_dims) == 0:
        # If the problem is fully continuous, fall back to standard optimization.
        return _optimize_acqf(
            opt_inputs=dataclasses.replace(
                opt_inputs,
                return_best_only=True,
            )
        )
    if not (
        isinstance(non_cont_dims, list)
        and len(set(non_cont_dims)) == len(non_cont_dims)
        and min(non_cont_dims) >= 0
        and max(non_cont_dims) <= dim - 1
    ):
        raise ValueError(
            "`discrete_dims` and `cat_dims` must be dictionaries with unique, disjoint "
            "integers as keys between 0 and num_dims - 1."
        )
    discrete_dims_t = torch.tensor(
        list(discrete_dims.keys()), dtype=torch.long, device=tkwargs["device"]
    )
    cat_dims_t = torch.tensor(
        list(cat_dims.keys()), dtype=torch.long, device=tkwargs["device"]
    )
    non_cont_dims = torch.tensor(
        non_cont_dims, dtype=torch.long, device=tkwargs["device"]
    )
    cont_dims = complement_indices_like(indices=non_cont_dims, d=dim)
    # Fixed features are all in cont_dims. Remove them, so they don't get optimized.
    ff_idcs = torch.tensor(
        list(fixed_features.keys()), dtype=torch.long, device=tkwargs["device"]
    )
    cont_dims = cont_dims[(cont_dims.unsqueeze(-1) != ff_idcs).all(dim=-1)]
    candidates = torch.empty(0, dim, **tkwargs)
    for _q in range(q):
        # Generate starting points.
        best_X, best_acq_val = generate_starting_points(
            opt_inputs=opt_inputs,
            discrete_dims=discrete_dims,
            cat_dims=cat_dims,
            cont_dims=cont_dims,
        )

        done = torch.zeros(len(best_X), dtype=torch.bool, device=tkwargs["device"])
        for _step in range(options.get("maxiter_alternating", MAX_ITER_ALTER)):
            starting_acq_val = best_acq_val.clone()
            best_X[~done], best_acq_val[~done] = discrete_step(
                opt_inputs=opt_inputs,
                discrete_dims=discrete_dims,
                cat_dims=cat_dims,
                current_x=best_X[~done],
            )

            best_X[~done], best_acq_val[~done] = continuous_step(
                opt_inputs=opt_inputs,
                discrete_dims=discrete_dims_t,
                cat_dims=cat_dims_t,
                current_x=best_X[~done],
            )

            improvement = best_acq_val - starting_acq_val
            done_now = improvement < options.get("tol", CONVERGENCE_TOL)
            done = done | done_now
            if done.float().mean() >= STOP_AFTER_SHARE_CONVERGED:
                break

        new_candidate = best_X[torch.argmax(best_acq_val)].unsqueeze(0)
        candidates = torch.cat([candidates, new_candidate], dim=-2)
        # Update pending points to include the new candidate.
        if q > 1:
            acq_function.set_X_pending(
                torch.cat([base_X_pending, candidates], dim=-2)
                if base_X_pending is not None
                else candidates
            )
    if q > 1:
        acq_function.set_X_pending(base_X_pending)

    if post_processing_func is not None:
        candidates = post_processing_func(candidates)

    with torch.no_grad():
        acq_value = acq_function(candidates)  # compute joint acquisition value
    return candidates, acq_value


def complement_indices_like(indices: Tensor, d: int) -> Tensor:
    r"""Computes a tensor of complement indices: {range(d) \\ indices}.
    Same as complement_indices but returns an integer tensor like indices.
    """
    return torch.tensor(
        complement_indices(indices.tolist(), d),
        device=indices.device,
        dtype=indices.dtype,
    )


def complement_indices(indices: list[int], d: int) -> list[int]:
    r"""Computes a list of complement indices: {range(d) \\ indices}.

    Args:
        indices: a list of integers.
        d: an integer dimension in which to compute the complement.

    Returns:
        A list of integer indices.
    """
    return sorted(set(range(d)).difference(indices))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Ensemble posteriors. Used in conjunction with ensemble models.
"""

from __future__ import annotations

import torch
from botorch.posteriors.posterior import Posterior
from torch import Tensor
from torch.distributions.multinomial import Multinomial


class EnsemblePosterior(Posterior):
    r"""Ensemble posterior, that should be used for ensemble models that compute
    eagerly a finite number of samples per X value as for example a deep ensemble
    or a random forest."""

    def __init__(self, values: Tensor, weights: Tensor | None = None) -> None:
        r"""
        Args:
            values: Values of the samples produced by this posterior as
                a `(b) x s x q x m` tensor where `m` is the output size of the
                model and `s` is the ensemble size.
            weights: Optional weights for the ensemble members as a tensor of shape
                `(s,)`. If None, uses uniform weights.
        """
        if values.ndim < 3:
            raise ValueError("Values has to be at least three-dimensional.")
        self.values = values
        self._weights = weights.to(values) if weights is not None else None
        # Pre-compute normalized weights and mixture properties for efficiency
        self._mixture_dims = list(range(self.values.ndim - 2))
        self._normalized_weights = self._compute_normalized_weights()
        self._normalized_mixture_weights = self._compute_normalized_mixture_weights()

    @property
    def ensemble_size(self) -> int:
        r"""The size of the ensemble"""
        return self.values.shape[-3]

    @property
    def mixture_size(self) -> int:
        r"""The total number of elements in the mixture dimensions"""
        return self.values.shape[:-2].numel()

    def _compute_normalized_weights(self) -> Tensor:
        r"""Compute and cache normalized weights."""
        if self._weights is not None:
            return self._weights / self._weights.sum(dim=-1, keepdim=True)
        else:
            return (
                torch.ones(
                    self.ensemble_size,
                    dtype=self.dtype,
                    device=self.device,
                )
                / self.ensemble_size
            )

    def _compute_normalized_mixture_weights(self) -> Tensor:
        r"""Compute and cache normalized mixture weights."""
        if self._weights is not None:
            unnorm_weights = self._weights.expand(self.values.shape[:-2])
            return unnorm_weights / unnorm_weights.sum(
                dim=self._mixture_dims, keepdim=True
            )
        else:
            return (
                torch.ones(
                    self.values.shape[:-2],
                    dtype=self.dtype,
                    device=self.device,
                )
                / self.mixture_size
            )

    @property
    def weights(self) -> Tensor:
        r"""The weights of the individual models in the ensemble.
        uniformly weighted by default."""
        return self._normalized_weights

    @property
    def mixture_weights(self) -> Tensor:
        r"""The weights of the individual models in the ensemble.
        uniformly weighted by default, and normalized over ensemble and
        batch dimensions of the model."""
        return self._normalized_mixture_weights

    @property
    def mixture_dims(self) -> list[int]:
        r"""The mixture dimensions of the posterior. For ensemble posteriors,
        this includes all dimensions except the last two (query points and outputs)."""
        return self._mixture_dims

    @property
    def device(self) -> torch.device:
        r"""The torch device of the posterior."""
        return self.values.device

    @property
    def dtype(self) -> torch.dtype:
        r"""The torch dtype of the posterior."""
        return self.values.dtype

    @property
    def mean(self) -> Tensor:
        r"""The mean of the posterior as a `(b) x n x m`-dim Tensor."""
        # Weighted average across ensemble dimension
        return (self.values * self.weights[..., None, None]).sum(dim=-3)

    @property
    def variance(self) -> Tensor:
        r"""The variance of the posterior as a `(b) x n x m`-dim Tensor.

        Computed as the weighted sample variance across the ensemble outputs.

        This treats weights as probability weights (normalized to sum to 1) and
        computes the unbiased weighted sample variance using the formula:
        Var = Σ(w_i * (x_i - μ)²) / (1 - Σw_i²)
        where the sum over w_i² is taken over the ensemble dimension only.
        Source: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean under
        "Reliability Weights".
        """
        if self.ensemble_size == 1:
            return torch.zeros_like(self.values.squeeze(-3))

        # Add dimensions for query points and outputs to enable broadcasting
        weights = self.weights[..., None, None]
        squared_deviations = (self.values - self.mean.unsqueeze(-3)) ** 2
        return (weights * squared_deviations).sum(dim=-3) / (1 - (weights**2).sum())

    @property
    def mixture_mean(self) -> Tensor:
        r"""The mixture mean of the posterior as a `(b) x n x m`-dim Tensor.

        Computed as the weighted average across the ensemble outputs.
        """
        return (self.values * self.mixture_weights[..., None, None]).sum(
            dim=self.mixture_dims
        )

    @property
    def mixture_variance(self) -> Tensor:
        r"""The mixture variance of the posterior as a `(b) x n x m`-dim Tensor.

        Computed as the weighted sample variance across the ensemble outputs.

        This treats weights as probability weights (normalized to sum to 1) and
        computes the unbiased weighted sample variance using the formula:
        Var = Σ(w_i * (x_i - μ)²) / (1 - Σw_i²) where w_i is normalized over the
        entire mixture, and the sum over w_i² is taken over all mixture dimensions.
        Source: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean under
        "Reliability Weights".
        """

        # Add dimensions for query points and outputs to enable broadcasting
        weights = self.mixture_weights[..., None, None]
        squared_deviations = (self.values - self.mixture_mean.unsqueeze(-3)) ** 2
        return (weights * squared_deviations).sum(dim=self.mixture_dims) / (
            1 - (weights**2).sum()
        )

    def _extended_shape(
        self,
        sample_shape: torch.Size = torch.Size(),  # noqa: B008
    ) -> torch.Size:
        r"""Returns the shape of the samples produced by the posterior with
        the given `sample_shape`.
        """
        return sample_shape + self.values.shape[:-3] + self.values.shape[-2:]

    @property
    def batch_shape(self) -> torch.Size:
        return self.values.shape[:-3]

    def rsample(
        self,
        sample_shape: torch.Size | None = None,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Based on the sample shape, base samples are generated and passed to
        `rsample_from_base_samples`.

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        if sample_shape is None or len(sample_shape) == 0:
            sample_shape = torch.Size([1])

        # NOTE This occasionally happens in Hypervolume evals when there
        # are no points which improve over the reference point. In this case, we
        # create a posterior for all the points which improve over the reference,
        # which is an empty set.
        if self.values.numel() == 0:
            return torch.empty(
                *self._extended_shape(sample_shape=sample_shape),
                device=self.device,
                dtype=self.dtype,
            )

        base_samples = (
            Multinomial(
                probs=self.mixture_weights,
            )
            .sample(sample_shape=sample_shape)
            .argmax(dim=-1)
        )
        return self.rsample_from_base_samples(
            sample_shape=sample_shape, base_samples=base_samples
        )

    def rsample_from_base_samples(
        self, sample_shape: torch.Size, base_samples: Tensor
    ) -> Tensor:
        r"""Sample from the posterior (with gradients) using base samples.

        This is intended to be used with a sampler that produces the corresponding base
        samples, and enables acquisition optimization via Sample Average Approximation.

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.
            base_samples: A Tensor of indices as base samples of shape
                `sample_shape`, typically obtained from `IndexSampler`.
                This is used for deterministic optimization. The predictions of
                the ensemble corresponding to the indices are then sampled.


        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        # Check that the first dimensions of base_samples match sample_shape
        if base_samples.shape != sample_shape + self.batch_shape:
            raise ValueError(
                f"Sample_shape={sample_shape + self.batch_shape} does not match "
                f"the leading dimensions of base_samples.shape={base_samples.shape}."
            )

        if self.batch_shape:
            # Values is always going to be 4-dimensional with this reshape,
            # even if we have more than one batch dimension
            values = self.values.reshape(
                ((self.batch_shape.numel(),) + self.values.shape[-3:])
            )

            # Collapse the base samples to enable index selecting along the
            # ensemble dim (dim -3)
            batch_numel = self.batch_shape.numel()
            collapsed_base_samples = base_samples.reshape(sample_shape + (batch_numel,))

            # First dimension is just 1, 2, 3, ..., batch_shape.numel() -1 to flatten
            # the first dimension and extract one index

            # second dimension extracts the ensemble member, for each element in the
            # entire batch shape
            return values[torch.arange(batch_numel), collapsed_base_samples].reshape(
                self._extended_shape(sample_shape=sample_shape)
            )
        return self.values[base_samples]
